/*---------------------------------------------------------------------

acquigen.cc

This file is part of CytoPacq

Copyright (C) 2007,2008 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Author: David Svoboda

Description: A module simulating signal acquisition using CCD detector. It performs e.g. sampling, noise addition, etc.

-------------------------------------------------------------------------*/

#include <i3d/transform.h>

#include "sim_macros.h"

#include "acquigen.h"
#include "mynoise.h"

using namespace i3d;
using namespace std;

//----------------------------------------------------------------------
void GetFinalResolution(const IniHandler &params, 
								Resolution &res)
{
	 // sample the image
	 float camera_pixel_size = params["camera"]["pixel size"];
	 float table_z_step = params["table"]["z step"];
	 float microscope_zoom = params["microscope"]["magnification"];
	 float microscope_adapter_zoom = 
				params["microscope adapter"]["magnification"];
	 float objective_zoom = params["objective"]["magnification"];
	 float cunit_zoom = params["confocal unit"]["magnification"];
	 float cunit_adapter_zoom = 
				params["confocal unit adapter"]["magnification"];

	 float fstep = camera_pixel_size / 
				(objective_zoom * 
				 microscope_zoom * 
				 microscope_adapter_zoom * 
				 cunit_zoom * 
				 cunit_adapter_zoom);

	 echo(" fstep: " << fstep);
	 echo(" table_z_step: " << table_z_step);

	 res = i3d::Vector3d<float>(1.0f/fstep, 
										 1.0f/fstep, 
										 1.0f/table_z_step);

	 echo(" final resolution: " << res.GetRes());
}

//----------------------------------------------------------------------
// SimulatePhotonDetection
//----------------------------------------------------------------------
template <class PHOTONS, class ELECTRONS>
void SimulatePhotonDetection(const IniHandler &params,
									  const Image3d<PHOTONS>& image_in,
									  Image3d<ELECTRONS> &image_out,
									  SamplingMode sampling = LANCZOS)
{
	 // sample the image
	 /*float camera_pixel_size = params["camera"]["pixel size"];
	 float table_z_step = params["table"]["z step"];
	 float microscope_zoom = params["microscope"]["magnification"];
	 float microscope_adapter_zoom = 
				params["microscope adapter"]["magnification"];
	 float objective_zoom = params["objective"]["magnification"];
	 float cunit_zoom = params["confocal unit"]["magnification"];
	 float cunit_adapter_zoom = 
				params["confocal unit adapter"]["magnification"];

	 float fstep = camera_pixel_size / 
				(objective_zoom * 
				 microscope_zoom * 
				 microscope_adapter_zoom * 
				 cunit_zoom * 
				 cunit_adapter_zoom);

	 Resolution fine_res(1.0f/fstep, 
								1.0f/fstep, 
								1.0f/table_z_step);

	 echo("\t iso res: " << fine_res.GetRes());
*/
	 Resolution fine_res;
	 GetFinalResolution(params, fine_res);

	 Image3d<PHOTONS> image_sampled(image_in);
    ResampleToDesiredResolution (image_sampled, fine_res, sampling);

	 // conversion of photon flux into electrons emitted
	 float dyn_usage = params["acquisition"]["dynamic range usage"];
	 int fwd = params["camera"]["full well depth"];

	 float max_capacity = fwd * dyn_usage / 100.0f;

	 image_out.CopyMetaData(image_sampled);
//	 PHOTONS max_ph_amount = image_in.GetVoxelData().max();
	 PHOTONS max_ph_amount = std::numeric_limits<PHOTONS>::max();
	 echo("max amount of photons: " << (size_t) image_in.GetVoxelData().max());
	 echo("max value in datatype:" << (size_t) max_ph_amount);

	 for (size_t i=0; i<image_sampled.GetImageSize(); i++)
	 {
		  float ph_amount = image_sampled.GetVoxel(i);
		  float el_amount = (ph_amount / max_ph_amount ) * max_capacity;

		  image_out.SetVoxel(i, (ELECTRONS) round(el_amount));
	 }

	 echo("max amount of electrons:" << (size_t) image_out.GetVoxelData().max());
}

//----------------------------------------------------------------------
// DigitizeSignal
//----------------------------------------------------------------------
template <class ELECTRONS, class ADU>
void SimulateADC(const IniHandler &params,
						  const Image3d<ELECTRONS> &image_in,
						  Image3d<ADU> &image_out)
{
	Image3d<ELECTRONS> image_with_noise(image_in);

	// Digitize - select the suitable ranges (minimum and maximum allowed
	// amount of electrons in the well) and digitize this range into
	// intensity (ADU) levels using the conversion factor 'ADC gain'
	float ADC_gain = params["camera"]["ADC gain"];
	int ADC_offset = params["camera"]["ADC offset"];

	ADU max_ADU = std::numeric_limits<ADU>::max();

	// quantize electrons into ADU:
	image_out.CopyMetaData(image_with_noise);
	for (size_t i=0; i<image_with_noise.GetImageSize(); i++)
	{
		 ELECTRONS el_amount = image_with_noise.GetVoxel(i);

		 float adu_amount_unbounded = 
					std::max(0, (int)el_amount - ADC_offset) / ADC_gain;

		 ADU adu_amount_bounded = 
					(adu_amount_unbounded > max_ADU) ?  
					max_ADU : 
					ADU(adu_amount_unbounded);

		 image_out.SetVoxel(i, adu_amount_bounded);
	}
}

//----------------------------------------------------------------------
// CallAcquigen
//----------------------------------------------------------------------
template <class PHOTONS, class ELECTRONS, class ADU>
void CallAcquigen(const IniHandler &params, 
						const Image3d<PHOTONS> &image_in,
						Image3d<ADU> &image_out,
						SamplingMode sampling)
{
	//--------------------------------------------------------------------
	// Sample the image to the resolution of CCD chip.
	// This corresponds to the acquisition process.
	//--------------------------------------------------------------------
	echo("photons detection & sampling");
	Image3d<ELECTRONS> image_el;
	SimulatePhotonDetection(params, image_in, image_el, sampling);

	//--------------------------------------------------------------------
	// generate Poisson noise
	//--------------------------------------------------------------------
	echo("adding Poisson noise");
	ApplyPoissonNoise(image_el);
	
	//--------------------------------------------------------------------
	// add dark current 
	//--------------------------------------------------------------------
	echo("adding dark current");
	float amount = params["camera"]["dark current"];
	float ac_time = params["acquisition"]["time"];
	AddDarkCurrent(image_el, amount, ac_time);

	//--------------------------------------------------------------------
	// add hot pixels
	//--------------------------------------------------------------------
	echo("adding hot pixels");
	int fwd = params["camera"]["full well depth"];
	for (size_t i=0; i<params["camera"].count("hot pixel"); i++)
	{
		 t_couple pos = params["camera"]("hot pixel",i);
		 SetAmountOfElectrons(image_el, 
									 (size_t) pos.x, 
									 (size_t) pos.y, 
									 (ELECTRONS) fwd);
	}

	//--------------------------------------------------------------------
	// add cold pixels
	//--------------------------------------------------------------------
	echo("adding cold pixels");
	for (size_t i=0; i<params["camera"].count("cold pixel"); i++)
	{
		 t_couple pos = params["camera"]("cold pixel",i);
		 SetAmountOfElectrons(image_el, 
									 (size_t) pos.x, 
									 (size_t) pos.y, 
									 (ELECTRONS) 0);
	}

	//--------------------------------------------------------------------
	// add baseline
	//--------------------------------------------------------------------
	echo("adding baseline level");
	int baseline = params["camera"]["baseline"];
	for (size_t i=0; i<image_el.GetImageSize(); i++)
	{
		 image_el.GetVoxelData()[i] += (ELECTRONS) baseline;
	}

	//--------------------------------------------------------------------
	// amplifier introducing read out noise
	//--------------------------------------------------------------------
	echo("simulating amplifier (adding read out noise)");
	int readout_noise = params["camera"]["read out noise"];
	ApplyReadoutNoise(image_el, (float) readout_noise);

	//--------------------------------------------------------------------
	// converting electrons to ADU (intensity levels)
	//--------------------------------------------------------------------
	echo("simulating ADC (digitizing the signal)");
	SimulateADC(params, image_el, image_out);
}

//---------------------------------------------------------------------------
// explicit instantiation - begin

template 
void CallAcquigen<GRAY16, size_t, GRAY16> (const IniHandler &, 
							  const Image3d<GRAY16> &, 
							  Image3d<GRAY16> &,
							  SamplingMode);

template  
void CallAcquigen<GRAY16, size_t, GRAY8>(const IniHandler &, 
							  const Image3d<GRAY16> &, 
							  Image3d<GRAY8> &,
							  SamplingMode);

template  
void CallAcquigen<GRAY8, size_t, GRAY8>(const IniHandler &, 
							  const Image3d<GRAY8> &, 
							  Image3d<GRAY8> &,
							  SamplingMode);

// explicit instantiation - end
//---------------------------------------------------------------------------
