/*---------------------------------------------------------------------

cytogen.cc

This file is part of CytoPacq

Copyright (C) 2007,2008 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Author: David Svoboda

Description: A module which creates an initial empty image. Afterwards, it fills it with user selected digital phantom(s). The definition of individual digital phantoms is implemented as a removable plugin.

-------------------------------------------------------------------------*/

#include <stdlib.h>
#include <time.h>

#include <i3d/resolution.h>
#include <i3d/regions.h>
#include <i3d/hrca_lut.h>
#include <i3d/voi.h>

#include "sim_macros.h"
#include "logs.h"

#include "plug.h"
#include "plugins/generic.h"

#include "cytogen.h"

/*****************************************************************************/

using namespace i3d;
using namespace std;

extern string warning_log_name;

/*****************************************************************************/

const char *phantom_section_name = "phantom";

/*****************************************************************************/
// Function that adds a new phantom into the image
/*****************************************************************************/

template <class T,class L> 
void CreateNewPhantom(Image3d<T> &img_phantoms, 
							 Image3d<L> &img_labels, 
							 const t_item_seq &phantom_properties,
							 const t_item_seq &common_properties,
							 const std::vector<std::string> &fnames)
{
	// Get the phantom name.
	const char *phantom_name = phantom_properties["name"]; 
	echo("looking for plugin for '" << phantom_name << "'");

	// Open the corresponding plugin.
	T_HANDLE plugin = FindAndOpenPlugin (phantom_name, fnames); 
				
	if (plugin == NULL) 
		 throw ERROR_DETAILS("Error opening plug-in.");

   // Read the necessary symbols from the plugin file.
   echo ("reading and verifying plugin symbols");

	// Load the functions defined in plug-ins
   t_create create = (t_create) GetSymbolFromPlugin (plugin, "create");

   if (create == NULL)
		 throw ERROR_DETAILS("Wrong plug-in. The symbol 'create' is missing.");

   t_destroy destroy = (t_destroy) GetSymbolFromPlugin (plugin, "destroy");

   if (destroy == NULL)
		 throw ERROR_DETAILS("Wrong plug-in. The symbol 'destroy' is missing.");

   // Create the phantom generator.
   phantom_generic *ph = create ();

   // Load the phantom parameters.
   ph->read_params (phantom_properties, common_properties);

	// Draw the phantom.
   echo ("new phantom [" << phantom_name << "] -- generation started ");
	
	typedef const char* (*t_string)(void);
	t_string id = (t_string) GetSymbolFromPlugin(plugin, "get_id");
	bool drawn;
  	
	if ((drawn = ph->draw(img_phantoms, img_labels)) == false)
	{
		 // If the position of phantom is fixed (defined by user)
		 // the new generation won't help. It is badly placed
		 if (phantom_properties.present("position"))
		 {
			  WarningLog(warning_log_name,
							 string("\nThe phantom '") + id() + 
							 "' cannot be placed in the\n" \
							 "position you defined. Skipping this phantom.\n");
		 }
		 // The phantom generation failed but it is randomly placed
		 // phantom. Hence try it again. May be you will have a luck 
		 // next time.
		 else 
		 {
			  echo("Generation failed, trying again."); 
			  
			  // generation failed -> try it once more
			  if ((drawn = ph->draw(img_phantoms, img_labels)) == false)
			  {
					// generation failed again
					if (img_labels.GetMaxValue() == L(0))
					{
						 // release the occupied memory
						 destroy(ph);
						 ReleasePlugin(plugin); 
						 
						 // the image is empty and nothing can be generated
						 // hence raise the exception
						 throw ERROR_DETAILS("Cannot generate new phantom.\n" \
											  " The volume of interest is too small.");
					}
					else
					{
						// the generation failed but image is not empty
						// (something has been generated before)
						// generate the warning only
						WarningLog(warning_log_name, 
								  string("\nA repetitive generation of phantom '") +
								  id() + "' failed\n"\
								  "due to lack of space in the image. "\
								  "Remedy: enlarge the VOI.\n");
					}
			  }
		 }
	}

	if (drawn)
	{
	   echo ("new phantom [" << phantom_name << "] -- generation completed ");
	}
	else
	{
	   echo ("new phantom [" << phantom_name << "] -- generation failed ");
	}

   // destroy the phantom generator
   destroy (ph);

   // close the plugin file
   ReleasePlugin(plugin);
}

/*****************************************************************************/
// Function that crops the image
/*****************************************************************************/
template <class T>
void CropImage(Image3d<T> &img, const VOI<PIXELS> &voi)
{
	 echo("input image size: " << img.GetSize());
	 echo("cropped size: " << voi);

	 Image3d<T> tmp(img, &voi);
	 img = tmp;
}

/*****************************************************************************/
// Function that pads the image
/*****************************************************************************/
template <class T>
void PadImage(Image3d<T> &img, 
				  const Vector3d<size_t> &pre_vector,
				  const Vector3d<size_t> &post_vector,
				  T value = T(0))
{
	 Vector3d<size_t> size = img.GetSize() + pre_vector + post_vector;
	 Vector3d<int> offset = - Vector3d<int>(pre_vector);
	 VOI<PIXELS> voi(offset, size);


	 Image3d<T> tmp;
	 tmp.CopyFromVOI(img, voi, true, value);

	 img = tmp;
}

/*****************************************************************************/
// Function that takes care of adding any object
/*****************************************************************************/

template <class T, class L> 
void CallCytogen(const IniHandler &params, 
					  Image3d<T> &img_phantoms, 
					  Image3d<L> &img_labels,
					  const char *plugin_directory)
{
	 // initialize random number generator
	 srand((unsigned)time(0));

	 // start generating the phantom(s)
	 echo("*** THE STAGE - \"BIOLOGICAL OBJECT GENERATION\" ***");

	 img_labels.CopyMetaData(img_phantoms);

	 // Check if there is some glass slide
	 size_t z_shift = 0;
	 if (params["specimen properties"].present("glass slide thickness"))
	 {
		  float thickness = 
					 params["specimen properties"]["glass slide thickness"];

		  z_shift = (size_t) floorf((thickness*img_phantoms.GetSizeZ()) / 100);
	 }

	 // Crop the volume of the image to prevent
	 // the images to be placed into the area occupied by glass.
	 if (z_shift > 0)
	 {
		  // the widht of the glass slide
		  Vector3d<size_t> sz = img_phantoms.GetSize();

		  if (z_shift >= sz.z)
		  {
				throw ERROR_DETAILS("Glass slide too thick.");
		  }

		  VOI<PIXELS> voi(0, 0, 0, sz.x, sz.y, sz.z-z_shift);

		  CropImage(img_phantoms, voi);
		  CropImage(img_labels, voi);
	 }

	 // search for all possible plugins
	 std::vector<std::string> plugin_files(0);
	 int num_of_plugins = SearchForPlugins(plugin_files, plugin_directory);

	 if (num_of_plugins == 0)
	 {
		  throw ERROR_DETAILS("No plugins found. Choose a suitable directory.");
	 }

	 echo("number of plugins: " << num_of_plugins);

	 // *** I ***
	 // First, add all the fixed objects. The user can define position 
	 // of some objects. This position must not collide with any random
	 // placed object. For this purpose the fixed objects are created first
	 echo("adding phantoms with fixed position");

	 for (size_t i=0; i<params.count(phantom_section_name); i++)
	 {
		  // focus only on objects with predefined position
		  if (params(phantom_section_name, i).present("position") == true)
		  {
				CreateNewPhantom(img_phantoms, 
									  img_labels, 
									  params(phantom_section_name, i), 
									  params["specimen properties"],
									  plugin_files);
		  }
	 }

	 // *** II ****
	 // Later, when all the fixed objects are ready and drawn into the image
	 // start handling with free objects
	 echo("adding phantoms with random position");

	 for (size_t i=0; i<params.count(phantom_section_name); i++)
	 {
		  // focus only on randomly placed objects
		  if (params(phantom_section_name, i).present("position") == false)
		  {
				CreateNewPhantom(img_phantoms, 
									  img_labels, 
									  params(phantom_section_name, i),
									  params["specimen properties"],
									  plugin_files);
		  }
	 }

	 echo("object addition finished");

	 // If there was some glass slide, pad the image to the original size.
	 // The part of the image, where the glass is located, is now empty.
	 if (z_shift > 0)
	 {
		  Vector3d<size_t> 
					 pre_pad(0,0,0),
					 post_pad(0,0,z_shift);

		  PadImage(img_phantoms, pre_pad, post_pad, T(0));
		  PadImage(img_labels, pre_pad, post_pad, L(0));
	 }

#ifdef DEBUG
	 echo("saving phantom image");
	 img_phantoms.SaveImage("debug-img-phantom.ics");
#endif

}

/***************************************************************************/

template <class T>
void CreateEmptyImage(const IniHandler &params, Image3d<T> &synth)
{
    // initialize random number generator
	 srand((unsigned)time(0));

	 // the beginning check out the basic system properties
	 float camera_pixel_size = params["camera"]["pixel size"];
	 t_couple camera_grid = params["camera"]["grid"];
	 t_couple voi_shift = params["voi"]["shift"];
	 t_triplet voi_size = params["voi"]["size"];

	 float microscope_zoom = 
				params["microscope"]["magnification"];

	 float microscope_adapter_zoom = 
				params["microscope adapter"]["magnification"];

	 float objective_zoom = 
				params["objective"]["magnification"];

	 float cunit_zoom = 
				params["confocal unit"]["magnification"];

	 float cunit_adapter_zoom = 
				params["confocal unit adapter"]["magnification"];

	 float precision = 
				params["subpixel precision"]["level"];

	 float fstep = camera_pixel_size / 
				(objective_zoom * 
				 microscope_zoom * 
				 microscope_adapter_zoom * 
				 cunit_zoom * 
				 cunit_adapter_zoom);

	 // set up the resolution (isotropic)
	 Resolution iso_res(precision/fstep, 
							  precision/fstep, 
							  precision/fstep);

	 // define USER AREA and CCD AREA
	 VOI<MICRONS> CCD_area_in_microns, user_area_in_microns;

	 // user volume (dimensions and offset)
	 user_area_in_microns.size.x = voi_size.x;
	 user_area_in_microns.size.y = voi_size.y;
	 user_area_in_microns.size.z = voi_size.z;
	 user_area_in_microns.offset.x = voi_shift.x;
	 user_area_in_microns.offset.y = voi_shift.y;
	 user_area_in_microns.offset.z = 0;

	 // CCD area (dimensions and zero offset)
	 CCD_area_in_microns.size.x = camera_grid.x * fstep;
	 CCD_area_in_microns.size.y = camera_grid.y * fstep;
	 CCD_area_in_microns.size.z = std::numeric_limits<float>::max(); 

	 // top-left corner of CCD is the reference point
	 CCD_area_in_microns.offset = Vector3d<float>(0);

	 // in debug more, put down some information
	 echo("camera pixel size: " << camera_pixel_size);
	 echo("fstep: " << fstep);
	 echo("voi size: " << user_area_in_microns.size);
	 echo("voi shift: " << user_area_in_microns.offset);
	 echo("CCD size: (" << CCD_area_in_microns.size.x << "," <<
				CCD_area_in_microns.size.y << ")");
	 echo("iso res: " << iso_res.GetRes());

	 // check whether the VOI is put inside CCD area
	 VOI<MICRONS> intersection_in_microns = 
				CCD_area_in_microns * user_area_in_microns;

	 VOI<PIXELS> 
				intersection_in_pixels(
							MicronsToPixels(intersection_in_microns.offset, iso_res),
						   MicronsToPixels(intersection_in_microns.size, iso_res)),
				user_area_in_pixels(
							MicronsToPixels(user_area_in_microns.offset, iso_res),
							MicronsToPixels(user_area_in_microns.size, iso_res));

	 echo("intersection (m): " << CCD_area_in_microns*user_area_in_microns);
	 echo("voi (m): " << user_area_in_microns);
	 echo("intersection (p): " << intersection_in_pixels);
	 echo("voi (p): " << user_area_in_pixels);

	 if (intersection_in_pixels != user_area_in_pixels)
	 {
		  throw ERROR_DETAILS("Your VOI is out of the camera domain."\
									 " Resize or shift the VOI.");
	 }

	 echo("image size (voxels): " << user_area_in_pixels.size);

	 // Create an empty image 
	 synth.MakeRoom(user_area_in_pixels.size);
	 synth.SetOffset(user_area_in_microns.offset);
	 synth.SetResolution(iso_res);
    synth.Clear();
}


//---------------------------------------------------------------------------
// explicit instantiation - begin

template void CallCytogen(const IniHandler&, 
								  Image3d<GRAY8>&, 
								  Image3d<LABEL_TYPE>&,
								  const char*);

template void CallCytogen(const IniHandler&, 
								  Image3d<GRAY16>&, 
								  Image3d<LABEL_TYPE>&,
								  const char*);

template void CreateEmptyImage(const IniHandler &, 
										 Image3d<GRAY8> &);

template void CreateEmptyImage(const IniHandler &, 
										 Image3d<GRAY16> &);

// explicit instantiation - end
//---------------------------------------------------------------------------
