/*---------------------------------------------------------------------

colon_tissue.cc

This file is part of CytoPacq

Copyright (C) 2009-2013 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Authors: 
- David Svoboda <svoboda@fi.muni.cz> August 2010
- Ondrej Homola <ondra.homola@gmail.com> June 2009


Description: A plug-in creating the digital phantom of healthy colon tissue.

-------------------------------------------------------------------------*/

#include <ctime>
#include <i3d/image3d.h>
#include <i3d/threshold.h>
#include <i3d/FastLevelSet.h>
#include <i3d/DistanceTransform.h>
#include <i3d/morphology.h>
#include <i3d/regions.h>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

#include "sim_macros.h"
#include "colon_tissue.h"
#include "src/diagram.h"
#include "src/add_shape.h"
#include "src/deformation.h"
#include "src/texture.h"
#include "src/myrandom.h"


//------------------------------------------------------------------------
// begin of some macros
//------------------------------------------------------------------------

#define RANDOM1 							(((double)rand())/RAND_MAX)
#define RANDOM_PLUS_MINUS(x) 			(-(x) + 2.0*(x)*RANDOM1)

// debug saving
#ifdef DEBUG
	#define DEBUG_IMAGE_SAVE(img, x) \
		echo("- saving debug image \""<< (x) <<"\""); \
		img.SaveImage( (x) ); \
		echo("- save successful.");
#else
	#define DEBUG_IMAGE_SAVE(img, x)
#endif

//------------------------------------------------------------------------
// end of some macros
//------------------------------------------------------------------------

using namespace i3d;
using namespace std;

//------------------------------------------------------------------------

extern "C" DLL_EXPORT const char *get_id()
{
	 return "Colon tissue";
}

//------------------------------------------------------------------------

bool phantom_colon_tissue::draw(i3d::Image3d<i3d::GRAY8>& img_phantoms, 
										  i3d::Image3d<LABEL_TYPE>& img_labels)
{
	 _create_shape(img_labels);

	 if (img_labels.GetMaxValue() == LABEL_TYPE(0))
		  return false;

	 _create_texture(img_labels, img_phantoms); 
	 _simulate_slicer(img_labels, img_phantoms);

	 return true;
}

//------------------------------------------------------------------------

bool phantom_colon_tissue::draw(i3d::Image3d<i3d::GRAY16>& img_phantoms, 
										  i3d::Image3d<LABEL_TYPE>& img_labels)
{
	 _create_shape(img_labels);

	 if (img_labels.GetMaxValue() == LABEL_TYPE(0))
		  return false;

	 _create_texture(img_labels, img_phantoms); 	
	 _simulate_slicer(img_labels, img_phantoms);

	 return true;
}

//-----------------------------------------------------------------------
	/**
	This method removes all cell and control points from the input sets which
	lie in a specified elliptic cylinder
	/param cell_points vector of cell points where the result will be stored.
	/param control_points vector of control points where the result 
	will be stored.
	/param centerX center of the ellipsoid
	/param centerY center of the ellipsoid
	/param angle angle of the ellipsoid
	/param torus_radiusX radius in x axis
	/param torus_radiusY radius in y axis
	/param torus_thickness increase of radius in ort. direction (not necessary)
	*/
void clear_elliptical_area(
			std::vector<Position> &cell_points, 
			std::vector<Position> &control_points,
			double centerX, 
			double centerY, 
			double angle, 
			double torus_radiusX, 
			double torus_radiusY, 
			double torus_thickness)
{
	//---- Erase all points in a small ellipse around the center --------------
	std::vector<Position> new_cell_points;
	std::vector<Position> new_control_points;

	double x, y, bx, by, a, b;

	a = ((torus_radiusX)*(torus_radiusX));
	b = ((torus_radiusY)*(torus_radiusY));

	for (size_t i = 0; i < control_points.size(); i++)
	{
		bx = control_points[i].x - centerX;
		by = control_points[i].y - centerY;
		x = cos(DEG_TO_RAD(angle)) * bx - sin(DEG_TO_RAD(angle)) * by;
		y = sin(DEG_TO_RAD(angle)) * bx + cos(DEG_TO_RAD(angle)) * by;

		if (x*x/a + y*y/b > 1.0)
		{
			new_control_points.push_back(control_points[i]);
		}
	}

	for (size_t i = 0; i < cell_points.size(); i++)
	{
		bx = cell_points[i].x - centerX;
		by = cell_points[i].y - centerY;
		x = cos(DEG_TO_RAD(angle)) * bx - sin(DEG_TO_RAD(angle)) * by;
		y = sin(DEG_TO_RAD(angle)) * bx + cos(DEG_TO_RAD(angle)) * by;

		if (x*x/a + y*y/b > 1.0)
		{
			new_cell_points.push_back(cell_points[i]);
		}
	}

	cell_points.assign(new_cell_points.begin(), new_cell_points.end());
	control_points.assign(new_control_points.begin(), new_control_points.end());
}

//----------------------------------------------------------------------
	/**			
	This method creates a model of a torus structure of cells using 
	the 3d voronoi diagram approach. The output is defined by two sets: 
	cell points and control points, that are to be passed to the
	get_diagram method.

	/param cell_points vector of cell points where the result will be stored.
	/param control_points vector of control points where the result 
	will be stored.
	/param centerX, 
	/param centerY, 
	/param centerZ, 
	/param angle, 
	/param shape_deformation_strength  Strength of extrema in the torus' 
	perimeter
	/param  shape_deformation_amount  number of extrema in the torus' perimeter
	/param  thickness_deformation_strength  strength of extrema in the torus' 
	thickness
	/param  thickness_deformation_amount  number of extrema in the torus' 
	thickness
	/param  density_deformation_strength  strength of extrema in the torus'
	cells density
	/param  density_deformation_amount  number of extrema in the torus'
	cells density
	/param torus_fuzziness  amount of individual cell's displacement from
	the elliptical perimeter
	/param torus_nodes  number of torus' cells
	/param torus_radiusX x size of the torus in microns
	/param torus_radiusY y size of the torus in microns
	/param torus_thickness approximate thickness of the torus in microns
	/param torus_height height of the torus in z coordinates (depth)
	/param torus_coning  determines how the torus's radius decreases with 
	its height
	*/
void generate_torus_cells(
	std::vector<Position> &cell_points, 
	std::vector<Position> &control_points,
	double centerX, 
	double centerY, 
	double centerZ, 
	double angle, 
	const int radius_variance_extrema,
	double radius_variance,
	const int thickness_variance_extrema,
	double thickness_variance,
	const int density_variance_extrema,
	double density_variance,
	double torus_cell_radial_displacement, 
	double cell_diameter, 
	double gaps,
	double torus_radiusX, 
	double torus_radiusY, 
	double torus_thickness, 
	double torus_height, 
	double torus_coning)
{
	// Generate torus radius irregularities
	valarray<double> radius_amplitudes(radius_variance_extrema);
	valarray<double> radius_offsets(radius_variance_extrema);
	
	for (int i = 0; i < radius_variance_extrema; i++)
	{
		radius_amplitudes[i] = (radius_variance/100) * RANDOM1;
		radius_offsets[i] = 360.0 * RANDOM1;		
	}

	// Generate torus thickness irregularities
	valarray<double> thickness_amplitudes(thickness_variance_extrema);
	valarray<double> thickness_offsets(thickness_variance_extrema);
	
	for (int i = 0; i < thickness_variance_extrema; i++)
	{
		thickness_amplitudes[i] = (thickness_variance/100) * RANDOM1;
		thickness_offsets[i] = 360.0 * RANDOM1;		
	}

	// Generate torus cell density irregularities
	valarray<double> density_amplitudes(density_variance_extrema);
	valarray<double> density_offsets(density_variance_extrema);
	
	for (int i = 0; i < density_variance_extrema; i++)
	{
		density_amplitudes[i] = density_variance * RANDOM1;
		density_offsets[i] = 360.0 * RANDOM1;		
	}

	// -------------------------------------------------------------------
	//                POLAR COORDINATES ITERATIVE GENERATION               
	// -------------------------------------------------------------------

	// Evaluate the theoretical volume of torus
	double torus_volume = M_PI*((torus_radiusX + torus_thickness)*
										 (torus_radiusY + torus_thickness) - 
										 torus_radiusX * torus_radiusY) * torus_height;

	// Lower the volume by gaps
	double real_torus_volume = torus_volume * (1-gaps/100);

	// Find out the volume of one cell nucleus
	double cell_volume = 4.0/3.0*M_PI*pow(cell_diameter/2.0, 3);

	// Find out the number of cell nuclei in one torus slice
	size_t num_of_cells_in_torus = (size_t) round(real_torus_volume / cell_volume);

	// basic iteration step - angle difference between two neighbouring cells
	const double step_size = 360.0/num_of_cells_in_torus;

	// offset has basic value 0, but is further modified by deformations
	double modified_step = 0.0;

	echo("torus volume: " << torus_volume);
	echo("torus volume without gaps: " << real_torus_volume);
	echo("estimated cell volume: " << cell_volume);
	echo("number of cells in torus: " << num_of_cells_in_torus);

	for(double i = 0; i < 360.0; i += modified_step)
	{
		// Deformation functions are constructed by sum of sine functions
		// This is not as nice as a regular Perlin noise (see literature) but 
		// it is sufficient for this purpose. 
		//
		// Why do we need such a complicated 'noise'? Why do we avoid the use of 
		// simply a Gaussian noise? We do not need a noise that flickers. We need 
		// a noise that form a continuous function. For example, if we need 
		// randomly generated curve, its nodes cannot jump within the space. Their
		// positions should be continuous in some sense.

		// Take care of local radius modifier - this modifier is responsible 
		// for the deformation of the torus radius in the i-th step
		double local_radius_modifier = 1.0 ;

		for (int j = 0; j < radius_variance_extrema; j++)
		{
			double s = pow(sin(DEG_TO_RAD(i - radius_offsets[j])), 3);
			local_radius_modifier -= std::max(0.0, radius_amplitudes[j] * s);
		}

		// Take care of local density modifier - this modifier is responsible
		// for the variable amount of cells (nuclei) in the local area
		double local_density_modifier = 1.0;

		for(int j = 0; j < density_variance_extrema; j++)
		{
			double s = pow(sin(DEG_TO_RAD(i - density_offsets[j])), 3);
			local_density_modifier += std::max(0.0, density_amplitudes[j] * s);			
		}

		modified_step = step_size * local_density_modifier;

		// Take care of local thickness modifier - this modifier is responsible
		// for the deformation of the torus thickness in the i-th step
		double local_thickness_modifier = 1.0 ;
		for (int j = 0; j < thickness_variance_extrema; j++)
		{
			double s = pow(sin(DEG_TO_RAD(i - thickness_offsets[j])), 3);
			local_thickness_modifier += std::max(0.0,   thickness_amplitudes[j] * s);
		}

		//----------------------------------------------------------------------
		//                            Cell point                                
		//----------------------------------------------------------------------
		// The position of each cell points will be compose of two parts:
		// (A) position - offset to ellipse (torus) center
		// (B) offset - distance from the ellipse border
		//
		// Point position on the ellipse border
		// (with respect to the ellipse center)
		double pos_on_elip_x = cos(DEG_TO_RAD(i)) * torus_radiusX;
		double pos_on_elip_y = sin(DEG_TO_RAD(i)) * torus_radiusY;

		// The exact position is modified by 'local_radius_modifier'
		double modif_pos_x = pos_on_elip_x * local_radius_modifier;
		double modif_pos_y = pos_on_elip_y * local_radius_modifier;
		double modif_pos_z = 0; // basic z-position is always equal to '0'

		// The position on the ellipse border is exactly defined by 'i' which
		// is actually the angle. In each such a position we can compute the 
		// normal vector which is perpendicular to the tangent in the given point.
		// So called 'thickness vector' is computed as a normal vector stretched by 
		// 'torus_thickness' and then modified by factor 'local_thickness_modifier'.
		double thick_x = cos(DEG_TO_RAD(i)) * torus_thickness;
		double thick_y = sin(DEG_TO_RAD(i)) * torus_thickness;

		double modif_thick_x = thick_x * local_thickness_modifier;
		double modif_thick_y = thick_y * local_thickness_modifier;
		
		// Each cell should be located inside the torus within the area defined
		// by thickness, i.e. by the vector (modif_thick_x, modif_thick_y).
		// The distribution within this area is driven by the factor
		// 'torus_cell_radial_displacement'.
		double displacement = RANDOM_PLUS_MINUS(torus_cell_radial_displacement);
		double cell_offset_x = modif_thick_x * displacement;
		double cell_offset_y = modif_thick_y * displacement;

		// 'z' offset is determined as a random position with respect to the
		// torus height. 
		double cell_offset_z_normalized = RANDOM_PLUS_MINUS(0.5);
		double cell_offset_z = cell_offset_z_normalized * torus_height;	

		// Basic 'x' and 'y' positions will be modifier by the coning effect
		// which depends on the 'z' position
		double coning_modifier = (1.0 + cell_offset_z_normalized * torus_coning);

		// We sum up basic cell position (modified by coning effect) and its
		// offset (distance from the basic ellipse border)
		double x = (modif_pos_x * coning_modifier) + cell_offset_x;	
		double y = (modif_pos_y * coning_modifier) + cell_offset_y;	
		double z = modif_pos_z + cell_offset_z;

		// Rotation of the whole torus point using rotation matrix
		double x_rotated = cos(DEG_TO_RAD(angle)) * x - sin(DEG_TO_RAD(angle)) * y;
		double y_rotated = sin(DEG_TO_RAD(angle)) * x + cos(DEG_TO_RAD(angle)) * y;

		// Transform the position into absolute coordinates
		double final_x = centerX + x_rotated,
			   final_y = centerY + y_rotated,
			   final_z = centerZ + z;

		cell_points.push_back(Position(final_x, final_y, final_z));
	
		//--------------------------------------------------------------------
		//                 Control point                                      
		//--------------------------------------------------------------------

		//--------- Border point above ----------
		// Place the nearest top control point at least at the top of the torus.
		control_points.push_back(Position(final_x, final_y, final_z - torus_height));

		//--------- Border point below ----------
		// Place the nearest bottom control point at least at the bottom of the torus.
		control_points.push_back(Position(final_x, final_y, final_z + torus_height));


		// ------- Border points on the left & on the right ---------
		// Take the current cell points and place the control points
		// INSIDE and OUTSIZE the torus. Do not place only two control points.
		// Build the whole sequence of control point pairs from the bottom of
		// torus to the top.
		for (double z_pos = -0.5 * torus_height; 
			  z_pos <= 0.5 * torus_height; 
			  z_pos +=  cell_diameter)
		{
			double z_normalized = z_pos / torus_height;
			coning_modifier = (1.0 + z_normalized * torus_coning);

			// Create the border point position OUTSIDE torus	
			x = (modif_pos_x * coning_modifier) + cell_offset_x;		
			y = (modif_pos_y * coning_modifier) + cell_offset_y;	

			// Outside, i.e. add the vector (modif_thick_x, modif_thick_y)
			x += modif_thick_x;		
			y += modif_thick_y;	

			// Rotation
			x_rotated = cos(DEG_TO_RAD(angle)) * x - sin(DEG_TO_RAD(angle)) * y;
			y_rotated = sin(DEG_TO_RAD(angle)) * x + cos(DEG_TO_RAD(angle)) * y;

			// Translation to the center
			final_x = centerX + x_rotated;
			final_y = centerY + y_rotated;
			final_z = centerZ + z_pos;

			// Store the OUTER control point.
			control_points.push_back(Position(final_x, final_y, final_z));

			// Create the border point position INSIDE torus	
			x = (modif_pos_x * coning_modifier) + cell_offset_x;		
			y = (modif_pos_y * coning_modifier) + cell_offset_y;

			// Inside, i.e. subtract the vector (modif_thick_x, modif_thick_y)
			x -= modif_thick_x;		
			y -= modif_thick_y;

			// Rotation
			x_rotated = cos(DEG_TO_RAD(angle)) * x - sin(DEG_TO_RAD(angle)) * y;
			y_rotated = sin(DEG_TO_RAD(angle)) * x + cos(DEG_TO_RAD(angle)) * y;

			// Translation to the center
			final_x = centerX + x_rotated;
			final_y = centerY + y_rotated;
			final_z = centerZ + z_pos;

			// Store the INNER control point.
			control_points.push_back(Position(final_x, final_y, final_z));

		} // loop - cylinder casing points

	} // loop - modified step

}

//-------------------------------------------------------------------------
	/**
	This method fills the specified voi with cells of certain size and density.
	/param cell_points vector of cell points where the result will be stored.
	/param control_points vector of control points where the result 
	will be stored.
	/param voi_width width of the area to be filled
	/param voi_height height of the area to be filled
	/param voi_depth depth of the area to be filled
	/param size average size of the cells to be created
	/param randomness variability of the cells size and position
	/param probability density of the generated cells
	*/
void generate_random_cells(
			std::vector<Position> &cell_points, 
			std::vector<Position> &control_points,
			double voi_width, 
			double voi_height, 
			double voi_depth, 
			double size, 
			double randomness, 
			double probability)
{

	if (size == 0 || probability == 0)
	{
		return;
	}

	// Generate border (control points) around the whole VOI:
	
	// two xy-planes (one in front of and one behind)
	for (double x = 0.0; x < voi_width; x += size )
	{
		for (double y = 0.0; y < voi_height; y += size )
		{
			control_points.push_back(Position(x, y, 0.0-size));
			control_points.push_back(Position(x, y, voi_depth+size));
		}
	}

	// two yz-planes
	for (double y = 0.0; y < voi_height; y += size )
	{
		for (double z = 0.0; z < voi_depth ; z += size )
		{
			control_points.push_back(Position(0.0-size, y,  z));
			control_points.push_back(Position(voi_width+size, y, z));
		}
	}

	// two xz-planes
	for (double x = 0.0; x < voi_width; x += size )
	{
		for (double z = 0.0; z < voi_depth; z += size )
		{
			control_points.push_back(Position(x, 0.0-size, z));
			control_points.push_back(Position(x, voi_height+size, z));
		}
	}

	// Place points in a uniform matrix with random offsets	
	for (double x = 0.0; x < voi_width; x += size)
	{
		for (double y = 0.0; y < voi_width; y += size)
		{
			for (double z = 0.0; z < voi_depth; z += size)
			{
				Position p(x + size * RANDOM_PLUS_MINUS(randomness), 
							  y + size * RANDOM_PLUS_MINUS(randomness), 
							  z + size * RANDOM_PLUS_MINUS(randomness));

				// only fraction of the generated points are cell points
				if (RANDOM1 < probability)
				{
					 cell_points.push_back(p);
				}
				else 
				{
					 control_points.push_back(p);
				}
			}
		}
	}
}
//------------------------------------------------------------------------
void phantom_colon_tissue::read_params(const t_item_seq &phantom_properties,
													const t_item_seq &comon_properties)
{
#ifdef DEBUG
    // read the object parameters from the input config file
	 surrounding_cells_density = 
				phantom_properties["surrounding cells density"];

	 surrounding_cell_randomness = 
				phantom_properties["surrounding cell randomness"];

	 cell_diameter = phantom_properties["cell diameter"];

	 cell_volume_variance = phantom_properties["cell volume variance"];

	 volume_of_gaps_in_torus = phantom_properties["volume of gaps in torus"];

	 torus_cell_radial_displacement = 
				phantom_properties["torus cell radial displacement"];

	 torus_skew = phantom_properties["torus skew"];

	 torus_coning = phantom_properties["torus coning"];

	 torus_radius = phantom_properties["torus radius"];

	 torus_radius_variance = phantom_properties["torus radius variance"];

	 torus_radius_variance_extrema = 
				phantom_properties["torus radius variance extrema"];

	 torus_thickness = phantom_properties["torus thickness"];

	 torus_thickness_variance = phantom_properties["torus thickness variance"];

	 torus_thickness_variance_extrema = 
				phantom_properties["torus thickness variance extrema"];

	 torus_density_variance = phantom_properties["torus density variance"];

	 torus_density_variance_extrema = 
				phantom_properties["torus density variance extrema"];

	 cell_deformation_level = phantom_properties["cell deformation level"];

	 cell_deformation_directions = 
				phantom_properties["cell deformation directions"];

	 cell_roundness = phantom_properties["cell roundness"];					

	 cell_surface_smoothness = phantom_properties["cell surface smoothness"];
	
	 cell_bright_border_thickness = 
				phantom_properties["cell bright border thickness"];

	 cell_brightness_attenuation = 
				phantom_properties["cell brightness attenuation"];

	 membrane_texture_grain_size = 
				phantom_properties["membrane texture grain size"];

	 membrane_texture_contrast = 
				phantom_properties["membrane texture contrast"];

	 membrane_texture_min_value = 
				phantom_properties["membrane texture min value"];
	
	 internal_texture_grain_size = 
				phantom_properties["internal texture grain size"];

	 internal_texture_variance = 
				phantom_properties["internal texture variance"];

	 internal_texture_contrast = 
				phantom_properties["internal texture contrast"];

	 internal_texture_min_value = 
				phantom_properties["internal texture min value"];

	 internal_texture_smoothness = 
				phantom_properties["internal texture smoothness"];

	 depth_of_cavity = phantom_properties["depth of cavity"];

	 texture_darkness = phantom_properties["texture darkness"];

	 texture_contrast = phantom_properties["texture contrast"];
	
	 texture_alpha = phantom_properties["texture alpha"];

	 texture_beta = phantom_properties["texture beta"];

	 texture_n = phantom_properties["texture n"];

    cell_brightness_variance = phantom_properties["cell brightness variance"];
	 
	 membrane_texture_weight = phantom_properties["membrane texture weight"];

	 internal_texture_weight = phantom_properties["internal texture weight"];

	 slice_thickness = phantom_properties["slice thickness"];

	 chromatin_granularity = phantom_properties["chromatin granularity"];
	 
#else
	 // use the predefined values
    surrounding_cells_density = 0.6f;
	 surrounding_cell_randomness = 0.7f;

	 cell_diameter = 6.3f;
	 cell_volume_variance = 90;

	 volume_of_gaps_in_torus = 40;
	 
	 torus_cell_radial_displacement = 0.4f;

	 torus_skew.x = 0.45f;
	 torus_skew.y = 0.45f;

	 torus_coning = 0.4f;

	 torus_radius.x = 36.0f;
	 torus_radius.y = 36.0f;

	 torus_radius_variance = 20;
	 torus_radius_variance_extrema = 6;

	 torus_thickness = 9.0f;
	 torus_thickness_variance = 60;
	 torus_thickness_variance_extrema = 6;

	 torus_density_variance = 1.2f;
	 torus_density_variance_extrema = 4;	 

	 cell_deformation_level = 0.3f;
	 cell_deformation_directions.x = -0.8f;
	 cell_deformation_directions.y = 0.2f;
	 
	 cell_roundness = 0.9f;					
	 cell_surface_smoothness = 25;
	
	 cell_bright_border_thickness = 0.8f;
	 cell_brightness_attenuation= 1.0f;

	 membrane_texture_grain_size = 6.5f; 
	 membrane_texture_min_value = 0.2f;
	 membrane_texture_contrast = 2.0f;
	
	 internal_texture_grain_size = 4.0f; 
	 internal_texture_variance = 70;
	 internal_texture_min_value = 0.25f;
	 internal_texture_contrast = 1.5f;
	 internal_texture_smoothness = 1.5f;
	 depth_of_cavity = 0.78f;

	 texture_darkness = 2.0f;
	 texture_contrast = 6.0f;

	 texture_alpha = 0.8f;
	 texture_beta = 0.7f;
	 texture_n = 6;

	 cell_brightness_variance = 85;

	 membrane_texture_weight = 1.1f; 
	 internal_texture_weight = 1.1f;

	 slice_thickness = 12.0f;	

	 chromatin_granularity = 7.0f;
	
#endif
}

//-------------------------------------------------------------------------
// This method creates a rough geometric model of the colon tissue
//-------------------------------------------------------------------------
void phantom_colon_tissue::_create_voronoi_regions(
			std::vector<Polyhedron> &list_of_cells)
{
	echo("Generating diagram");

	VoronoiDiagram voronoi;

	// initialize
	list_of_cells.clear();

	// initialize the random generator
	srand((unsigned)time(0)); 
	
	// ----- Surrounding points initialization -----
	// We create a separate Diagram object which will capture 
	// random cells around the tori
	std::vector<Position> surround_cell_points(0), surround_control_points(0);

	// This fills the point vectors, 
	// which are the input for the 3d voronoi calculation
	generate_random_cells(surround_cell_points, 
								 surround_control_points,
								 voi.x, voi.y, voi.z,
								 cell_diameter, 
								 surrounding_cell_randomness, 
								 surrounding_cells_density);

	// ----- Torus construction -----
	std::vector<Position> torus_centers(0);
	bool filled = false;
	int attempts = 0;
	bool collides = false;

	// If the VOI is too small that placing the torus center inside the VOI would
	// cause all the cell appearing outside the VOI, we purposely place the initial center
	// out of the VOI.
	if ((voi.x < torus_radius.x) || (voi.y < torus_radius.y))
		 torus_centers.push_back(Position(2*voi.x, 2*voi.y, 0));

	// We randomly determine centers of the tori.
	// Simply try placing another torus on a random position and see if it 
	// would not be too close to the others.
	// If no suitable position can be found after a couple of times, stop.
	while (!filled)
	{
		attempts++;
		Position new_point(RANDOM1 * voi.x, RANDOM1 * voi.y, 0);
		collides = false;

		for (size_t i = 0; i < torus_centers.size(); i++)
		{
			double dx = new_point.x - torus_centers[i].x;
			double dy = new_point.y - torus_centers[i].y;
			double dist = sqrt(dx*dx + dy*dy);
			if (dist < (torus_radius.x + torus_radius.y))
			{
				collides = true;
				break;
			}
		}

		if (!collides)
		{
			torus_centers.push_back(new_point);
			attempts = 0;
		}
		
		if (attempts > 255)
		{			
			filled = true;
		}
	}

	// Create each torus independently
	for (size_t i = 0; i < torus_centers.size(); i++)
	{
		double angle = 180.0*((double)rand())/RAND_MAX;
		
		// Make sure the space inside the ellipsoid is clear 
		// (erase all present points)
		clear_elliptical_area(surround_cell_points, 
									 surround_control_points,
									 torus_centers[i].x, 
									 torus_centers[i].y,
									 angle, 
									 torus_radius.x, 
									 torus_radius.y, 
									 torus_thickness);

		// points vectors (input for 3d voronoi)
		std::vector<Position> torus_cell_points, torus_control_points;

		torus_cell_points.clear();
		torus_control_points.clear();

		// Generate torus points (fill the control points vector)
		generate_torus_cells( 
							torus_cell_points, 
							torus_control_points,
							torus_centers[i].x, 
							torus_centers[i].y, 
							voi.z/2, 
							angle,
							torus_radius_variance_extrema,
							torus_radius_variance,
							torus_thickness_variance_extrema,
							torus_thickness_variance,
							torus_density_variance_extrema,
							torus_density_variance,
							torus_cell_radial_displacement,
							cell_diameter,
							volume_of_gaps_in_torus,
							torus_radius.x, 
							torus_radius.y, 
							torus_thickness, 
							voi.z, 
							torus_coning);

		// The point vectors for the torus and the transformation 
		// are passed to the generator, which builds the vorornoi diagram
		// and applies the transformation to the result.
		echo("torus number: " << i+1 );
		echo("torus cell points: " << torus_cell_points.size());
		echo("torus control points: " << torus_control_points.size());

		voronoi.Create(torus_cell_points, torus_control_points);

		const std::vector<Polyhedron>& 
				  td = voronoi.GetSkewedRegions(torus_skew.x, 
														  torus_skew.y, 
														  angle);

		echo("Number of regions: " << torus_cell_points.size());

		// The result is copied to the list of polyhedrons.
		// Use only the first 'torus_cell_points.size()' voronoi regions.
		for (size_t i = 0; i < torus_cell_points.size(); i++)
		{
			 if (td[i].GetNumberOfPlanes() > 3)
			 {
				  list_of_cells.push_back(td[i]);
			 }
		}

		torus_cell_points.clear();
		torus_control_points.clear();
	} // end loop (tori)

	// As soon as the torus cells are generated we can start generating
	// the cells that are randomly placed within the space

	echo("surround cell points: " << surround_cell_points.size());
	echo("surround control points: " << surround_control_points.size());

	voronoi.Create(surround_cell_points, surround_control_points);

	const std::vector<Polyhedron>& sd = voronoi.GetRegions();

	// The result is copied to the list of polyhedrons.
	// Use only the first 'surround_cell_points.size()' voronoi regions
	for(size_t i = 0; i < surround_cell_points.size(); i++)
	{
		 if (sd[i].GetNumberOfPlanes() > 3)
		 {
			  list_of_cells.push_back(sd[i]);
		 }
	}
	
	echo("Diagram completed");
}

//-------------------------------------------------------------------------
// This function crops the front and back z-planes (like the slicer)
// from the final phantom image
//-------------------------------------------------------------------------
template <class T, class L> void 
phantom_colon_tissue::_simulate_slicer(i3d::Image3d<L> &img_labels,
													i3d::Image3d<T> &img_phantoms)
{
	float res_in_z = img_labels.GetResolution().GetRes().z;
	float max_depth = img_labels.GetSizeZ() / res_in_z;

	L last_label = img_labels.GetMaxValue();

	for (size_t i=0; i<img_labels.GetImageSize(); i++)
	{
		 if (img_labels.GetVoxel(i) == last_label)
		 { 
			  size_t depth_in_pixels = img_labels.GetZ(i);
			  float depth_in_microns = depth_in_pixels / res_in_z; 
			  
			  if (fabs(max_depth/2.0f - depth_in_microns) > slice_thickness/2.0f)
			  {
					img_labels.SetVoxel(i, L(0));
					img_phantoms.SetVoxel(i, T(0));
			  }
		 }
	}
}

/**************************************************************************
 *
 * Create Shape
 *
 ************************************************************************/
template <class L> void
phantom_colon_tissue::_create_shape(i3d::Image3d<L> &img_labels)
{
   voi = PixelsToMicrons(img_labels.GetSize(),
								 img_labels.GetResolution()); 

	std::vector<Polyhedron> list_of_nuclei;

	_create_voronoi_regions(list_of_nuclei);

	echo("Creating image mask...");
	Image3d<bool> mask; 
	mask.CopyMetaData(img_labels);

	// Real size of the image VOI - um = micrometers
	Vector3d<float> size_um = PixelsToMicrons(img_labels.GetSize(),
															img_labels.GetResolution());

	// Generate cells and merges them with the mask
	for (unsigned i = 0; i < list_of_nuclei.size(); i++)
	{
		// echo("*** Voronoi region *** " << i << " volume: " << list_of_nuclei[i].GetVolume());

		// Remove cells which are COMPLETELY out of the VOI (image domain),
		// the diagram can be bigger that the VOI.
		const Polyhedron& region = list_of_nuclei[i];
		if ((region.GetLowBoundingPoint().x < 0 || 
			  region.GetTopBoundingPoint().x > size_um.x) ||
			 (region.GetLowBoundingPoint().y < 0 || 
			  region.GetTopBoundingPoint().y > size_um.y) ||
			 (region.GetLowBoundingPoint().z < 0 || 
			  region.GetTopBoundingPoint().z > size_um.z))
		{
			// echo("Region " << i << " is out of visible area -> rejected");
			continue;
		}
		
		// ---------- creation of a cell image ----------
		Image3d<bool> cell_image;
		cell_image.SetResolution(img_labels.GetResolution());
		
		Coords pos;
		
		// Does the cell overlap with some already existing object?
		if (!_create_cell_image(cell_image, pos, region, mask, img_labels))
		{
			// echo("Region " << i << " overlaps another one -> rejected");
			continue;
		}

		// ---------- Merging the Cell image with the Mask image ---------
		Coords abs_pos;

		for (size_t x = 0; x < cell_image.GetSizeX(); x++)
		{
			 abs_pos.x = pos.x + x; 
			 
			 for (size_t y = 0; y < cell_image.GetSizeY(); y++)
			 {
				  abs_pos.y = pos.y + y;
				  
				  for (size_t z = 0; z < cell_image.GetSizeZ(); z++)
				  {
						abs_pos.z = pos.z + z; 
					
						// This test prevents from accessing the positions
						// outside the image domain + it copies only nonzero
						// pixels to the image mask
						if (cell_image.GetVoxel(x, y, z) == true &&
							 mask.Include(abs_pos.x, abs_pos.y, abs_pos.z))
						{																		
							 mask.SetVoxel(abs_pos.x, abs_pos.y, abs_pos.z, true);	
						}
				  } // z-loop
			 } // y-loop
		} // x-loop
	} // lost_of_nuclei-loop

	echo("Done creating image mask.");

	DEBUG_IMAGE_SAVE(mask, "debug00_base_mask.ics");

	//------------------------ MASK DEFORMATION ------------------------------
	echo("Performing deformation of the mask...");	

	float res_x = img_labels.GetResolution().GetRes().x;

	// create the speed image
	i3d::Image3d<float> speed_img;
	speed_img.CopyMetaData(img_labels);

	GenerateSpeedImage(speed_img, 
							 cell_diameter * cell_roundness,
							 cell_surface_smoothness,
							 cell_deformation_directions.x,
							 cell_deformation_directions.y);

	//speed_img.SaveImage("speed.ics");

	// define the deformation domain
	i3d::Image3d<bool> free_space;
	Threshold(img_labels, free_space, L(0), L(0));

	// perform the deformation
	Deformation(mask, &free_space, speed_img, 
					cell_deformation_level*res_x);

	echo("Deformation completed.");	

	DEBUG_IMAGE_SAVE(mask, "debug01_deformed_mask.ics");	

	AddNewLabel(img_labels, mask, true);

	DEBUG_IMAGE_SAVE(img_labels, "debug01_deformed_mask_pridana.ics");	
}

/**************************************************************************
 *
 * Create Texture
 *
 ************************************************************************/
template <class T, class L>  
void phantom_colon_tissue::_create_texture(
		const i3d::Image3d<L> &img_labels,
		i3d::Image3d<T> &image_phantoms)
{
	// Activate the random generator
	gsl_rng_env_setup();
	gsl_rng *r = gsl_rng_alloc(gsl_rng_default);

	echo("Generating texture...");

	// ----- Invert the mask => store to distance map ---
	// First, we copy out the mask of the actual digital phantom. It should
	// be the component with the highest label value in 'img_labels'.
	Image3d<bool> colon_mask;
	Threshold(img_labels, 
				 colon_mask, 
				 (L)(img_labels.GetMaxValue()));

	// Label individual nuclei within this colon tissue phantom.
	LabeledImage3d<L,bool> colon_labels;
	colon_labels.CreateRegionsFF(colon_mask);

	Image3d<float> accumulator;
	accumulator.CopyMetaData(img_labels);
	accumulator.Clear();

	for (L index=1; index<=colon_labels.get_max_label(); index++)
	{
		 if (colon_labels.components[index].voxel_value == false)
		 {
			  echo("background detected -- skipping this component");
			  continue;
		 }

		 echo("creating texture for the nucleus: " <<
				(size_t) index << " of " << 
				(size_t) colon_labels.get_max_label());

		 // fetch the nucleus that corresponds to the label 'index'
		 // and generate the texture only for this nucleus
		 Image3d<bool> nucleus_mask;
		 Image3d<T> nucleus_texture;

		 colon_labels.GetComponentMask(index, nucleus_mask);
		 nucleus_texture.CopyMetaData(nucleus_mask);
		 nucleus_texture.Clear();

		 //--------------- DISTANCE MAP TEXTURE COMPONENT --------------
		 Image3d<float> distance_map;
		 distance_map.CopyMetaData(nucleus_mask);
		 echo("distance map");

		 // Prepare the mask image for the generation of the distance map.
		 for(size_t i = 0; i <nucleus_mask.GetImageSize(); i++)
		 {
			  if (nucleus_mask.GetVoxel(i) == false)
			  { 
					distance_map.SetVoxel(i, 1.0f);
			  }
			  else 
			  {
					distance_map.SetVoxel(i, 0.0f);
			  }
		 }
	
		 // Generate the distance map (with the Euclidean metric)
		 EDM(distance_map);

		 // ----------------- MEMBRANE BASE ----------------
		 echo("membrane base");
		 // Find out the maximum value (within the distance map) 
		 float max_value = cell_diameter/2;

		 for (size_t i=0; i<distance_map.GetImageSize(); i++)
		 {
			  max_value = std::max(max_value, distance_map.GetVoxel(i));
		 }

		 // Invert the distance map in the voxels that belong
		 // to the foreground (nucleus) + apply some weighting
		 // using 'cell_brightness_attenuation' factor
		 for (size_t i=0; i<distance_map.GetImageSize(); i++)
		 {
			  if (nucleus_mask.GetVoxel(i) == true)
			  {
					float value = distance_map.GetVoxel(i);
					float inverted_value = max_value - value;
					float normalized_value = inverted_value / max_value;
					float weighted_value = pow(normalized_value, 
														cell_brightness_attenuation);
					distance_map.SetVoxel(i, weighted_value);
			  }
		 } 

//		 distance_map.SaveImage("debug02_membrane_base.ics");
		 
		 // Now the base for nucleus membrane is created. 
		 // It is too regular however.
		 // In the following steps, we will modify it.

		 //------- MEMBRANE MODIFIER --------------------------------
		 echo("membrane modifier");
		 Image3d<float> membrane_img;
		 membrane_img.CopyMetaData(nucleus_mask); 
		 
		 Image3d<float> membrane_modifier;
		 membrane_modifier.CopyMetaData(nucleus_mask); 
		 
		 // The membrane modifier will be a Perlin noise.
		 DoPerlin3D(membrane_modifier,
				cell_diameter/membrane_texture_grain_size,
				texture_alpha,
				texture_beta,
				texture_n); 
		 
		 Stretch(membrane_modifier, membrane_texture_min_value, 1.0f); 
//		 membrane_modifier.SaveImage("debug03_membrane_modifier.ics");
		 
		 // Membrane base is modified by the Perlin noise. 
		 // The membrane is simply
		 // point-wise multiplied by Perlin noise.
		 for (size_t i=0; i<distance_map.GetImageSize(); i++)
		 {
			  float value = distance_map.GetVoxel(i);
			  value *= membrane_modifier.GetVoxel(i);
			  membrane_img.SetVoxel(i, value);
		 } 
		 
		 // Make the image more contrasty. This prevents the final image 
		 // (after convolution) to be too blurred/smooth.
		 IncreaseContrast(membrane_img, membrane_texture_contrast);
//		 membrane_img.SaveImage("debug04_membrane.ics");

		 //------ INTERNAL STRUCTURE TEXTURE COMPONENT---------------------
		 echo("internal structure");
		 Image3d<float> internal_texture_img;
		 internal_texture_img.CopyMetaData(nucleus_mask); 
		 
		 // The internal texture will be formed of Perlin waves.
		 float texture_granularity = 
					GenFromRange(internal_texture_grain_size *
									 (1.0f - internal_texture_variance/100.0f),
									 internal_texture_grain_size *
									 (1.0f + internal_texture_variance/100.0f));

		 echo("texture: " << texture_granularity);

		 DoPerlin3D(internal_texture_img, 
				  cell_diameter/texture_granularity, 
				  texture_alpha*internal_texture_smoothness, 
				  texture_beta*internal_texture_smoothness, 
				  texture_n);
	
		 // The texture will contain the values in the range 
		 // 'internal_texture_min_value' -> 1.0
		 Stretch(internal_texture_img, internal_texture_min_value, 1.0f);

		 // the texture must be of high contrast. 
		 // If it were of low contrast, after convolution
		 // with PSF we would get blurred image.
		 IncreaseContrast(internal_texture_img, internal_texture_contrast);

//		 internal_texture_img.SaveImage("debug05_internal.ics");

		 // ---------------- MERGING TEXTURES ----------------------------------
		 echo("merging textures");
		 // Final voxel value is computed as WEIGHTED MAX of membrane texture
		 // and internal texture.
 		 float int_weight = internal_texture_weight/
				  (membrane_texture_weight + internal_texture_weight);

		 float surf_weight = membrane_texture_weight/
				  (membrane_texture_weight + internal_texture_weight);

		 Stretch(distance_map, depth_of_cavity, 1.0f);

		 Image3d<float> final_texture_img;
		 final_texture_img.CopyMetaData(nucleus_mask);
		 final_texture_img.Clear(); 
		 
		 if (membrane_texture_weight + internal_texture_weight > 0.0)
		 {			
			  for (size_t i = 0; i <internal_texture_img.GetImageSize(); i++)
			  {	
					if (nucleus_mask.GetVoxel(i) == true)
					{
						 float intern = internal_texture_img.GetVoxel(i) * int_weight;
						 float surface = membrane_img.GetVoxel(i) * surf_weight;

						 // Moreover the internal texture is weighted by 
						 // the distance from the
						 // center of the nucleus. In general, the centre of each 
						 // nucleus is dark
						 // (low level of staining).
						 intern *= distance_map.GetVoxel(i);

						 float val = std::max(intern, surface);

						 final_texture_img.SetVoxel(i, val);				
					}
			  }	
		 }

		 // normalize the image
		 Stretch(final_texture_img, 0.0f, 1.0f);
	
		 for (size_t i=0; i<final_texture_img.GetImageSize(); i++)
		 {
			  float value = final_texture_img.GetVoxel(i);
			  value = pow(value, texture_darkness);
			  final_texture_img.SetVoxel(i, value);
		 }

		 IncreaseContrast(final_texture_img, texture_contrast);
//		 final_texture_img.SaveImage("debug06_final_texture.ics");

		 echo("uneven staining");

		 // Use the Gaussian distribution to generate 
		 // the non-uniformity of staining
		 // the individual nuclei. In general, 
		 // not all the cells are stained equally.
		 double factor =  1.0f +
					gsl_ran_gaussian(r, cell_brightness_variance/100.0f) + 1.0f; 
		 
		 // Track the individual nuclei (their pixels) and modify their brightness.
		 for (size_t i=0; i<final_texture_img.GetImageSize(); i++)
		 {
			  float fvalue = final_texture_img.GetVoxel(i); 
			  
			  if (fvalue > 0.0f)
			  {
					final_texture_img.SetVoxel(i, (float) (fvalue * factor));
			  }
		 }

		 //final_texture_img.SaveImage("debug07_final_texture_stained.ics");
		 StickToImage(accumulator, final_texture_img, nucleus_mask);
	}

//	DEBUG_IMAGE_SAVE(accumulator, "debug07_accumulator.ics");	

	// this code makes the images to seem grainy
	Image3d<float> zrna;
	zrna.CopyMetaData(accumulator);

	DoPerlin3D(zrna, 
				  cell_diameter/(internal_texture_grain_size*4.5), 
				  texture_alpha*internal_texture_smoothness, 
				  texture_beta*internal_texture_smoothness, 
				  texture_n);

	Stretch(zrna, 0.0f, 1.0f);
	IncreaseContrast(zrna, chromatin_granularity); 

	/*
#ifdef DEBUG
	zrna.SaveImage("zrna.ics");
#endif
*/

	for (size_t i=0; i<accumulator.GetImageSize(); i++)
	{
		 float value = zrna.GetVoxel(i) * accumulator.GetVoxel(i);
		 accumulator.SetVoxel(i, value);
	}

	// TODO: photobleaching bude treba jeste poradne dostudovat, ale
	// je to prave tento efekt - nektere snimky jsou tmavsi a nektere 
	// svetlejsi -> nektere se vic vysvitily. Ted to necham takto, ale
	// bude to treba doprogrmovat obecne do vsech fantomu!
   float photobleaching = 50.0f;
   float range = GenFromRange(1.0f - photobleaching/100.0f,
	 								   1.0f + photobleaching/100.0f);

	// The image does not cover the whole range (max == 65K).
	Stretch(accumulator, 0.0f, (float) 33000 * range);

   Image3d<T> accumulator_gray;
	FloatToGrayNoWeight(accumulator, accumulator_gray);

	// Stick the result to the output image 'image_phantoms' among other already
	// generated phantoms.
	StickToImage(image_phantoms, accumulator_gray, colon_mask);

	echo("Done creating texture.");
}


/****************************************************************************
* This method rasterizes a single cell c into cell_image
* Performs interior/exterior test for each voxel in the cell's bounding volume
* Performs rounding of the cell using Gaussian blur with thresholding
* The output image will not extend behind the bordering cell structure 
* (no voxels are outside the polyhedron)
* Returns position of the cell image's front top left corner voxel 
* coordinates in the mask image
***************************************************************************/
template <class L> bool 
phantom_colon_tissue::_create_cell_image(Image3d<bool>& cell_image, 
													  Coords &cell_position,
													  const Polyhedron& region, 
													  const Image3d<bool>& mask,
													  const Image3d<L> &img_labels)
{
	// define one pixel wide border
	const int border = 1;

	Vector3d<float> res = cell_image.GetResolution().GetRes();

	Position top = region.GetTopBoundingPoint();
	Position low = region.GetLowBoundingPoint();

	Vector3d<size_t> cell_size(
		(size_t) round((low.x - top.x) * res.x) + border*2,
		(size_t) round((low.y - top.y) * res.y) + border*2,
		(size_t) round((low.z - top.z) * res.z) + border*2);

	double cell_volume = region.GetVolume();
	const double mean_cell_volume = M_PI*powf(cell_diameter, 3.0f)/6.0;

	double allowed_variance = cell_volume * cell_volume_variance / 100.0;
	
	if (abs(mean_cell_volume - cell_volume) > allowed_variance)
	{
		 // echo("Too large or too small -> rejected");
		 return false;
	}
	
	// Determine the cell's position (top left corner) in pixels
	cell_position = Coords((int)( round((top.x) * res.x)) - border,
								  (int)( round((top.y) * res.y)) - border,
								  (int)( round((top.z) * res.z)) - border);

	cell_image.MakeRoom(cell_size);

	// ------------------- Fill the cell's area -----------------	
	
	for (size_t i=0; i<cell_image.GetImageSize(); i++)
	{
		 // relative position within the region
		 Vector3d<size_t> rel_pos_pix = cell_image.GetPos(i);
		 // absolute position with respect to the origin [0,0,0]
		 Coords abs_pos_pix = cell_position + rel_pos_pix;
		 // absolute position (in microns)
		 Position pos_um = PixelsToMicrons(abs_pos_pix, res); 
			 
	    // test whether the voxel is located inside the region
		 if (region.contains(pos_um))
		 { 
			  // avoid touching the voxels outside the image mask
			  if (mask.Include(abs_pos_pix))
			  {
					// test whether the regions does (not) overlap
					if ((mask.GetVoxel(abs_pos_pix) == true) ||
						 (img_labels.GetVoxel(abs_pos_pix) != L(0)))
					{
						 return false;
					}
					
			  } 
			  
			  cell_image.SetVoxel(rel_pos_pix, true);
		 }				
	}

	// Number '3' is precision level of this fast level set method. Setting
	// this parameter higher brings higher precision in evaluation but also
	// slows down the performance. Number '3' was found (by the developer
	// of fast level set method - Martin Maska) to be a good option.
	//
	// If the 'cell_roundness' is equal to zero the individual nuclei
	// will be pressed to each other in such a way that nobody will be
	// able to distinguish the individual nuclei. For this purpore, if
	// 'cell_roundness == 0' we will apply one 3D erosion.

	if (cell_roundness == 0.0)
	{ 
		 Image3d<bool> temp = cell_image;
		 Erosion(temp, cell_image, nb3D_26);
	}
	else
	{
		 i3d::fls::Routine<int> *routine = 
					i3d::fls::MCMRoutine(cell_image, NULL, 3);

		 routine -> Initialize(); 
	
		 // Surface of a cuboid
		 size_t surface = (cell_image.GetSizeX() * cell_image.GetSizeY() +
								 cell_image.GetSizeY() * cell_image.GetSizeZ() +
								 cell_image.GetSizeZ() * cell_image.GetSizeX()) * 2; 
		 
		 // Empirically tested that one half (cell_roundess == 0.5) of the image 
		 // surface is a suitable number representing the numer of iterations.
		 routine -> Iterate((size_t)round(surface * cell_roundness)); 
		 routine -> Mask(cell_image.begin(), i3d::fls::ResultCreator::InteriorAndInterface);
		 delete routine;
	}

	return true;
}

//------------------------------------------------------------------------

extern "C" DLL_EXPORT phantom_colon_tissue* create()
{	
	 phantom_colon_tissue *ph = new phantom_colon_tissue;
	 return ph;
}

//------------------------------------------------------------------------

extern "C" DLL_EXPORT void destroy(phantom_colon_tissue* ph)
{
	 delete ph;
}
//------------------------------------------------------------------------
