/*---------------------------------------------------------------------

colon_tissue.h

This file is part of CytoPacq

Copyright (C) 2009-2013 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Authors: 
- David Svoboda <svoboda@fi.muni.cz> August 2010
- Ondrej Homola <ondra.homola@gmail.com> June 2009


Description: A plugin creating the digital phantom of healthy colon tissue.

-------------------------------------------------------------------------*/

#ifndef __PHANTOM_COLON_TISSUE__
#define __PHANTOM_COLON_TISSUE__

#include <vector>
#include <i3d/image3d.h>

#include "generic.h"
#include "src/diagram.h"

//------------------------------------------------------------------------

class phantom_colon_tissue : public phantom_generic
{
private:
  // Size vector of the created diagram (in microns)
  i3d::Vector3d<double> voi;

  // Density of the cells around the tori. Real values from 0.0 to 1.0, 
  // where zero means no cells.
  float surrounding_cells_density;

  // Average diameter of cells (in microns) 
  float cell_diameter;

  // Cell volume variance (in %)
  float cell_volume_variance;

  // Randomness of surrounding cells' position.
  // Recommended values are in the range <0.0; 1.0>
  float surrounding_cell_randomness;

  // Approximate number of torus' cells. It also determines cells' size, 
  // according to the torus' diameter.
  //int torus_nodes;
 
  // How much of the torus volume is not occupied (in %)
  float volume_of_gaps_in_torus;

  // the torus' radius in microns
  t_couple torus_radius;

  // Approximate thickness (size) of the torus cells in microns. 
  // Thickness affects cells' size only in the direction "from the center". 
  float torus_thickness;

  // Strength of extrema in the torus' thickness. Amount of thin and thick 
  // areas along the diameter.
  // Recommended values from 0.0 to 2.0.   
  float torus_thickness_variance;

  // number of extrema in the torus' thickness 
  // Recommended values from 1 to 10
  int torus_thickness_variance_extrema;

  // Maximum value of individual cell's displacement from the elliptical 
  // perimeter. Treated as a weight which is multiplied with diameter.
  // The reasonable values are from 0.0 to 1.0
  float torus_cell_radial_displacement;

  // Amount of cells' surface deformation. Used in Deformation algorithm.
  float cell_deformation_level;
  t_couple cell_deformation_directions;

  // Parameter that drives the smoothness of the generated objects.
  // The values are <0;oo>. Zero means no smoothing.
  // Infinity means absolute smoothing. However, setting this value to infinity
  // will cause the collapse of an object into a point which finally vanishes.
  // The optimal value of this parameter is 0.5
  float cell_roundness;
  float cell_surface_smoothness;

  // torus' skew vector in micrometers. Offset between 
  // the top and bottom torus centers.
  t_couple torus_skew;

  // Determines how the torus's radius decreases with it's height. 
  // Recommended values from 0.0 to 1.0.
  float torus_coning;

  // Strength of extrema in the torus' perimeter. The depth of big bumps 
  // on the torus' diameter.
  // Recommended values from 0.0 to 2.0.
  float torus_radius_variance;

  // number of extrema in the torus' perimeter
  // Recommended values from 0 to 10
  int torus_radius_variance_extrema;

  // strength of extrema in the torus' cells density. Amount of dense areas 
  // with small cells and sparse areas with large cells..
  // Recommended values from 0.0 to 2.0.
  float torus_density_variance;

  // number of extrema in the torus' cells density
  // Recommended values from 1 to 10.
  int torus_density_variance_extrema;

  // Thickness of a "brighter" layer on the cell's surface in micrometers
  float cell_bright_border_thickness;

  // Weight of the texture in the outer radius of the cell
  // Recommended values from 1.0 to 5.0
  float membrane_texture_weight;

  // Weight of the texture in the center of the cell
  // Recommended values from 1.0 to 5.0.
  float internal_texture_weight;

  // Speed of texture attenuation towards the center of the cell
  // Recommended values from 1.0 to 5.0
  float cell_brightness_attenuation;

  // Membrane texture parameters
  float membrane_texture_grain_size;
  float membrane_texture_contrast;
  float membrane_texture_min_value;

  // Internal texture parameters
  float internal_texture_grain_size;
  float internal_texture_variance;
  float internal_texture_contrast;
  float internal_texture_min_value;
  float internal_texture_smoothness;
  float depth_of_cavity;

  // Final texture parameters
  float texture_darkness;
  float texture_contrast;

  // Brightness of the cell
  float cell_brightness_variance;

  // Generic texture (Perlin noise) parameters
  // passed to DoPerlin3D function
  float texture_alpha;
  float texture_beta;
  int texture_n;

  // How much the chromatin appear grainy instead of homogeneous
  // (larger number -> more intense effect)
  float chromatin_granularity ;

  // The depth of slice
  float slice_thickness;

	// ---------- internal methods --------------------
	 template <class L>
	 bool _create_cell_image(i3d::Image3d<bool>&,
									 i3d::Coords &,
									 const Polyhedron&, 
									 const i3d::Image3d<bool>&,
									 const i3d::Image3d<L> &);

	 void _create_voronoi_regions(std::vector<Polyhedron> &);
			 
	 template <class L>
	 void _create_shape(i3d::Image3d<L> &);

	 template <class T, class L>
	 void _create_texture(const i3d::Image3d<L> &, 
								 i3d::Image3d<T> &);

	 template <class T, class L>
	 void _simulate_slicer(i3d::Image3d<L> &,
								  i3d::Image3d<T> &);


  public:                               
	 // -------------- public methods -------------------------
	 virtual void read_params(const t_item_seq &,
									  const t_item_seq &);

	 virtual bool draw(i3d::Image3d<i3d::GRAY8>&, 
							 i3d::Image3d<LABEL_TYPE>&);

	 virtual bool draw(i3d::Image3d<i3d::GRAY16>&, 
							 i3d::Image3d<LABEL_TYPE>&);

	 /// virtual destructor
	 virtual ~phantom_colon_tissue() {};
};

//------------------------------------------------------------------------


#endif // __PHANTOM_COLON_TISSUE__
