/*---------------------------------------------------------------------

generic.h

This file is part of CytoPacq

Copyright (C) 2007-2013 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Author: David Svoboda

Description: A header file containing the definition of abstract class that creates the general digital phantom.

-------------------------------------------------------------------------*/

#ifndef _PHANTOM_GENERIC_
#define _PHANTOM_GENERIC_

#include <stdio.h>
#include <typeinfo>
#include <iostream>
#include <vector>

#include <i3d/image3d.h>

#include "ini/iniparser.h"
#include "../cytogen.h"

#ifdef _WIN32
	#define DLL_EXPORT __declspec(dllexport)
#else
	#define DLL_EXPORT
#endif

//------------------------------------------------------------------------
/** An abstract phantom class does nothing.
 *
 * The inherited classes should be responsible for correct reading of
 * the input parameters and creating of the phantom.
 */
class phantom_generic
{ 
  protected:
			 float photobleaching;

  public:
			 /// constructor - empty
			 phantom_generic() : 
						photobleaching(0.0f) {};

			 /// destructor 
			 virtual ~phantom_generic() {};

			 /** This method gets the pointer to INI structure (stored in
			  * the memory) and saves it for further usage.
			  */
			 virtual void read_params(const t_item_seq &phantom_properties,
											  const t_item_seq &common_properties);

			 /** The method puts the phantom and its mask to the given images.
			  * The parameter position (if given) defines the location of 
			  * the generated object (digital phantom). 
			  */
			 virtual bool draw(i3d::Image3d<i3d::GRAY8> &img_phantoms,
									 i3d::Image3d<LABEL_TYPE> &img_labels) = 0;

			 virtual bool draw(i3d::Image3d<i3d::GRAY16> &img_phantoms,
									 i3d::Image3d<LABEL_TYPE> &img_labels) = 0;
};

//------------------------------------------------------------------------
// some important function datatypes:
//------------------------------------------------------------------------

typedef phantom_generic* (*t_create)(void);
typedef void (*t_destroy) (phantom_generic*);

//------------------------------------------------------------------------

class phantom_with_position: public phantom_generic
{
  protected:
			 i3d::Vector3d<float> *position;
			 float glass_slide_thickness;
			 bool influence_of_gravity;
			 float cluster_effect;

  public:
			 /// constructor - empty
			 phantom_with_position() : 
						position(NULL),
						glass_slide_thickness(0.0f),
						influence_of_gravity(false),
						cluster_effect(0.0f) {};

			 /** This method gets the pointer to INI structure (stored in
			  * the memory) and saves it for further usage.
			  */
			 virtual void read_params(const t_item_seq &phantom_properties,
											  const t_item_seq &common_properties);

			 /// destructor 
			 virtual ~phantom_with_position();
};

//------------------------------------------------------------------------

#endif
