/*---------------------------------------------------------------------

granulocyte.cc

This file is part of CytoPacq

Copyright (C) 2007-2013 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Author: David Svoboda

Description: A plugin creating the digital phantom of granulocyte nucleus.

-------------------------------------------------------------------------*/

#include <list>

#include "i3d/neighbours.h"
#include "i3d/morphology.h"
#include "i3d/draw.h"
#include "i3d/threshold.h"
#include "i3d/DistanceTransform.h"

#include "sim_macros.h"

#include "src/myrandom.h"
#include "src/add_shape.h"
#include "src/deformation.h"
#include "src/texture.h"
#include "src/collisions.h"

#include "granulocyte.h"

#define sqr(x) \
		  ((x)*(x))

#define dist2(v1,v2) \
		  (sqr((v1).x-(v2).x) + sqr((v1).y-(v2).y) + sqr((v1).z-(v2).z))

using namespace i3d;

//------------------------------------------------------------------------
extern "C" DLL_EXPORT const char *get_id()
{
	 return "Granulocyte nucleus";
}

//------------------------------------------------------------------------
// get the vector of 3D point and sort them to make the shortest sequence
// using greedy travelling salesman algorithm
//------------------------------------------------------------------------

void _reorder_nodes(std::vector<Vector3d<float>/**/> &nodes)
{
	 // initial values
	 float max_sum = 0.0f;
	 size_t node_index_with_max_sum = 0;

	 // find the furthest point:
	 for (size_t i=0; i<nodes.size(); i++)
	 {
		  float curr_sum = 0;
		  for (size_t j=0; j<nodes.size(); j++)
		  {
				curr_sum += dist2(nodes[i], nodes[j]);
		  }

		  if (curr_sum > max_sum)
		  {
				max_sum = curr_sum;
				node_index_with_max_sum = i;
		  }
	 }

	 std::vector<Vector3d<float>/**/> ordered_nodes(0);
	 ordered_nodes.push_back(nodes[node_index_with_max_sum]);

	 // move the nodes (except the furthest one) into the temporary list
	 std::list<Vector3d<float>/**/> bag_of_nodes;
	 std::list<Vector3d<float>/**/>::iterator it, min_it;

	 for (size_t i=0; i<nodes.size(); i++)
	 {
		  if (i != node_index_with_max_sum)
				bag_of_nodes.push_back(nodes[i]);
	 }

	 // Start from this furthest point and add (step by step) the following
	 // point. Each new point should be the nearest point to the last added.
	 while (!bag_of_nodes.empty())
	 {
		  Vector3d<float> last_node = bag_of_nodes.back();

		  float min_dist = std::numeric_limits<float>::max();
		  
		  for (it = bag_of_nodes.begin(); it != bag_of_nodes.end(); it++)
		  {
				float d = dist2(last_node, *it);

				if (min_dist > d)
				{
					 min_dist = d;
					 min_it = it;
				}
		  }

		  ordered_nodes.push_back(*min_it);
		  bag_of_nodes.erase(min_it);
	 }

	 // copy the result to the output
	 nodes = ordered_nodes;
}

//------------------------------------------------------------------------
void phantom_granulocyte::read_params(const t_item_seq &phantom_properties,
												  const t_item_seq &common_properties)
{
	 phantom_with_position::read_params(phantom_properties,
													common_properties);

#ifdef DEBUG
  d_mean = phantom_properties["diameter"];
  d_variance = phantom_properties["diameter variance"];
  spread = phantom_properties["spread of lobes"];
  num_of_lobes = phantom_properties["number of lobes"];
  lobe_size_variance = phantom_properties["lobe size variance"];
  lobe_irregularity = phantom_properties["lobe shape irregularity"];
  snake_width = phantom_properties["channel width"];
  shell_thickness = phantom_properties["shell thickness"];
  chromatin_irregularity = phantom_properties["chromatin irregularity"];
#else
  d_mean = 6.0f;
  d_variance = 10;
  spread = 2.0f;
  num_of_lobes.x = 3;
  num_of_lobes.y = 5;
  lobe_size_variance = 25;
  lobe_irregularity = 1.0f;
  shell_thickness = 0.8f;
  snake_width = 0.2f;
  chromatin_irregularity = 0.7f;
#endif
}

//------------------------------------------------------------------------

bool phantom_granulocyte::draw(i3d::Image3d<i3d::GRAY8>& img_phantoms, 
										 i3d::Image3d<LABEL_TYPE>& img_labels)
{
	 if (!_create_shape(img_labels))
		  return false; 
	 
	 _create_texture(img_labels, img_phantoms); 
	 return true;
}

//------------------------------------------------------------------------

bool phantom_granulocyte::draw(i3d::Image3d<i3d::GRAY16>& img_phantoms, 
										 i3d::Image3d<LABEL_TYPE>& img_labels)
{
	 if (!_create_shape(img_labels))
		  return false; 
	 
	 _create_texture(img_labels, img_phantoms); 
	 return true;
}



//------------------------------------------------------------------------
template <class T> void phantom_granulocyte
::_imitate_chromatin_structure(Image3d<T> &img_texture, 
										 const Image3d<bool> &img_mask)
{
  // make the temporary Image3d<float> and compute the distance map
  Image3d <float> img_dist;
  img_dist.CopyMetaData(img_mask);
  for (size_t i=0; i<img_mask.GetImageSize(); i++)
  {
		img_dist.SetVoxel(i, img_mask.GetVoxel(i));
  }

  EDM (img_dist, 0, img_dist.GetVoxelData().max(), true);

#ifdef DEBUG
  img_dist.SaveImage("debug-gran-distance.ics");
#endif

  // all the position which bear value larger than 'shell thickness'
  // are set the 'shell thickness' level
  for (size_t i=0; i<img_dist.GetImageSize(); i++)
  {
		if (img_dist.GetVoxel(i) > shell_thickness)
		{
			 img_dist.SetVoxel(i, shell_thickness);
		}
  }

  // normalize and invert the area fits the object mask
  float dt_max = img_dist.GetVoxelData ().max (); 

  for (size_t i = 0; i < img_dist.GetImageSize (); i++)
    {
      if (img_mask.GetVoxel (i) != false)
        {
				float fvalue = img_dist.GetVoxel(i);
				float new_fvalue = 1.0f - (fvalue/dt_max);

				img_dist.SetVoxel(i, new_fvalue);
		  }
    }

  // add Perlin noise to ensure the distance map is not so uniform
  AddPerlin (img_dist, dt_max / chromatin_irregularity);

#ifdef DEBUG
  img_dist.SaveImage("debug-gran-distance-affected.ics");
#endif

  img_texture.CopyMetaData(img_mask);

  // combine the results and put them to the phantom image
  for (size_t i = 0; i < img_texture.GetImageSize (); i++)
    {
      if (img_mask.GetVoxel (i) == true)
        {
				// get to original value
				// T old_value = img_texture.GetVoxel(i);
				T old_value = std::numeric_limits<T>::max();

				// get texture value
				float texture = img_dist.GetVoxel(i);

				// multiply the texture with the image of the object
				// (pixelwise)
				float new_value = (float)old_value * texture;
				img_texture.SetVoxel(i, (T) round(new_value));
		  }
    }

#ifdef DEBUG
  img_texture.SaveImage("debug-gran-texture.ics");
#endif

}

//------------------------------------------------------------------------
template <class L>
bool phantom_granulocyte::_create_shape(i3d::Image3d<L>& img_labels)
{
  std::vector<Vector3d<float>/**/> vector_of_positions(0);

  float nucleus_d = GenNum (d_mean, d_variance);
  float nucleus_r = nucleus_d/2;
  Vector3d<float> 
			 mean, // mean position
			 dev(spread*nucleus_r/3, // standard deviation
				  spread*nucleus_r/3, 
				  spread*nucleus_r/3); 

  Image3d<bool> occupied;
  Threshold(img_labels, occupied, L(1), std::numeric_limits<L>::max());

  // (A) localization of the centre
  // initially, try to find a suitable location for the whole nucleus
  if (position == NULL)
  {
		Vector3d<float> dims(nucleus_d);
		Vector3d<size_t> sz = MicronsToPixels(dims, img_labels.GetResolution()); 
	
		Image3d<bool> initial;
		initial.MakeRoom(sz + 2); // plus some boundary
		initial.SetResolution(img_labels.GetResolution());
		initial.Clear(); 
		
		AddEllipsoid(initial, 
						 nucleus_r, nucleus_r, nucleus_r,
						 nucleus_r, nucleus_r, nucleus_r); 
		
		Vector3d<float> *rand_mean = 
				  find_position_in_cluster(initial, 
													occupied, 
													cluster_effect, 
													influence_of_gravity);

	   if (rand_mean == NULL)
			 return false;

		mean = *rand_mean;

		delete rand_mean;
  }
  else
  {
		mean = *position;
  }
	
  // (B) generation of lobes
  // how many lobes should be generated?
  int lower = (int) num_of_lobes.x,
		upper = (int) num_of_lobes.y;
  size_t count = lower + (rand() % (upper-lower+1));
  //size_t count = (size_t) round(GenFromRange(num_of_lobes.x, num_of_lobes.y));
  echo("\t -> number of lobes: " << count);

  // specify the mean lobe diameter
  float lobe_d = nucleus_d / powf((float)count, 1.0f/3.0f);
  echo("\t -> mean lobe diameter: " << lobe_d);

  Image3d<bool> accumulator;
  accumulator.CopyMetaData(img_labels);
  accumulator.Clear();

  // generate all the lobes
  for (size_t i=0; i<count; i++)
  {
		echo("-- adding lobe number: " << (i + 1));

		Vector3d<float> lobe_pos, lobe_r;

		lobe_r.x = GenNum (lobe_d/2, lobe_size_variance);
	  	lobe_r.y = GenNum (lobe_d/2, lobe_size_variance);
  		lobe_r.z = GenNum (lobe_d/2, lobe_size_variance);
		float max_r = std::max(lobe_r.x, std::max(lobe_r.y, lobe_r.z));
		Vector3d<float> centre(max_r), dims = centre * 2.0f;
		Vector3d<size_t> sz = MicronsToPixels(dims, img_labels.GetResolution());

		Image3d<bool> initial;
		initial.MakeRoom(sz);
		initial.SetResolution(img_labels.GetResolution());
		initial.Clear();

  		float rotation_xy = GenFromRange (-90, +90),
				rotation_xz = GenFromRange (-90, +90);

		AddEllipsoid(initial,
						 centre.x, centre.y, centre.z,
						 lobe_r.x, lobe_r.y, lobe_r.z,
						 rotation_xy, rotation_xz);

	   if (initial.GetVoxelData().max() == false)
	   {
			 throw "The object you want you generate is smaller than one voxel.\n"
					" Solution: Increase the resolution.";
		}


		Vector3d<float> *random_pos;

	   echo("\t adding lobe");
		random_pos = find_position_gaussian_weighted(initial, occupied, 
																	mean, dev, false);

		if (random_pos == NULL)
			  throw "No more space for new granulocyte lobe \n"
							 "Try to enlarge you VOI.";

		echo("\t position found");

		MergeBinaryImages(occupied, initial, *random_pos - centre);
		MergeBinaryImages(accumulator, initial, *random_pos - centre);
		echo("\t solution merged with previous objects");

		vector_of_positions.push_back(*random_pos);

		delete random_pos;
  } // loop - adding the lobes


  // deform the granulocyte
  Image3d<float> img_speed;
  img_speed.CopyMetaData(img_labels);
  echo ("generating speed image");
  GenerateSpeedImage(img_speed, nucleus_r/2);

  Image3d<bool> free_space;
  echo("generating domain");
  Threshold(img_labels, free_space, L(0), L(0));

#ifdef DEBUG
  {
		accumulator.SaveImage("debug-gran-nondeformed.ics");
		free_space.SaveImage("debug-gran-domain.ics");
		img_speed.SaveImage("debug-gran-speed.ics");
  }
#endif

  echo ("deformation in progress");
  Deformation(accumulator, &free_space, img_speed, lobe_irregularity);
  echo ("deformation finished");

#ifdef DEBUG
  {
		Image3d<GRAY8> help;
		help.CopyMetaData(accumulator);
		for (size_t i=0; i<accumulator.GetImageSize(); i++)
		{
			 if (accumulator.GetVoxel(i) == true) 
				  help.SetVoxel(i, GRAY8(255));
		}
		help.SaveImage("debug-gran-deformed.ics");
  }
#endif

  // sort the vector with points (each one is the centre of some lobe)
  _reorder_nodes(vector_of_positions);

  // now, connect the lobes with snakes
  // TODO: neresi se prekryvy!!!!
  for (size_t i=0; i<(vector_of_positions.size()-1); i++)
  {
	 Vector3d<float> 
			from_pos = vector_of_positions[i],
			to_pos = vector_of_positions[i+1];

	 Vector3d<float> mid_pos = from_pos+to_pos;
	 mid_pos /= 2;
	 mid_pos.x += 1.0f;

    std::vector<Vector3d<float>/**/> nodes;
	 nodes.clear();
	 nodes.push_back(from_pos);
	 nodes.push_back(mid_pos);
	 nodes.push_back(to_pos);
  	 AddSnake(accumulator, nodes, snake_width, true);
  }

  return AddNewLabel(img_labels, accumulator);
}

//------------------------------------------------------------------------
template <class T, class L> void phantom_granulocyte::
_create_texture(const i3d::Image3d<L> &img_labels,
					 i3d::Image3d<T> &img_phantoms)
{
	 L lbl = img_labels.GetMaxValue();

	 if (lbl == (L) 0) // there is no object, do nothing
		  return;

	 Image3d<bool> mask;
    Image3d<T> texture;

	 GetLabelMask(img_labels, lbl, mask);

  	 _imitate_chromatin_structure(texture, mask);

    // copy the new object into the image with phantoms
	 StickToImage(img_phantoms, texture, mask);
}

//------------------------------------------------------------------------

extern "C" DLL_EXPORT phantom_granulocyte* create()
{
	 phantom_granulocyte* ph = new phantom_granulocyte;
	 return ph;
}

//------------------------------------------------------------------------

extern "C" DLL_EXPORT void destroy(phantom_granulocyte* ph)
{
	 delete ph;
}
//------------------------------------------------------------------------
