/*---------------------------------------------------------------------

add_shape.cc

This file is part of CytoPacq

Copyright (C) 2007-2013 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Author: David Svoboda

Description: The functions that can: merge binary images, add some primitives (ellipsoids, cylinders, snakes).

-------------------------------------------------------------------------*/

#include <iostream>
#include <math.h>

#include <gsl/gsl_errno.h>
#include <gsl/gsl_spline.h>

#include <i3d/DistanceTransform.h>

#include "sim_macros.h"
#include "add_shape.h"

using namespace std;
using namespace i3d;

/*************************************************************************/

template <class T> bool AddNewLabel(Image3d<T> &img,
												const Image3d<bool>& mask,
												bool ignore_overlaps)
{
	 if (img.GetResolution().GetRes() != mask.GetResolution().GetRes())
		  throw ERROR_DETAILS("Resolutions do not match.");

	 if (mask.GetMaxValue() == false) // nothing to add
	 {
		  echo("exiting -- nothing to add");
		  return false;
	 }

	 Offset diff_offset = mask.GetOffset() - img.GetOffset();
	 Vector3d<size_t> pos = MicronsToPixels(diff_offset, img.GetResolution());

	 size_t x,y,z;
	 bool fit = true;

	 // First, check if it is possible to include new phantom.
	 for (size_t i=0; i<mask.GetImageSize() && fit; i++)
	 {
		  if (mask.GetVoxel(i))
		  {
				x = pos.x + mask.GetX(i);
				y = pos.y + mask.GetY(i);
				z = pos.z + mask.GetZ(i);

				// vymazat - begin
				if (!img.Include(x,y,z))
					 echo("JE MIMO");

				if ((!ignore_overlaps) && (img.OnBorder(x,y,z)))
					 echo("JE NA HRANE");

				if  ((!ignore_overlaps) && (img.GetVoxel(x,y,z) != T(0)))
					 echo("PREPISUJE JINY");
				// vymazat - end

				if ((!img.Include(x,y,z)) || 
					 ((!ignore_overlaps) && img.OnBorder(x,y,z)) ||
					 ((!ignore_overlaps) && (img.GetVoxel(x,y,z) != T(0))))
				{
					 fit = false;
				}
		  }
	 }

	 // If the placement is possible, do it.
	 if (fit == true)
	 {
		  T new_label = img.GetMaxValue() + 1;

		  for (size_t i=0; i<mask.GetImageSize(); i++)
		  {
				if (mask.GetVoxel(i))
				{
					 x = pos.x + mask.GetX(i);
					 y = pos.y + mask.GetY(i);
					 z = pos.z + mask.GetZ(i); 
					 
					 img.SetVoxel(x,y,z,new_label);
				}
		  }
	 }

	 echo("Final decision about placement: " << fit);
	 return fit;
}

/*************************************************************************/

template <class T> void GetLabelMask(const Image3d<T> &img,
												 T label,
												 Image3d<bool> &mask)
{
	 size_t x,y,z;
	 size_t min_x, min_y, min_z, max_x, max_y, max_z;

	 // initialize the variables
	 min_x = img.GetSizeX();
	 min_y = img.GetSizeY();
	 min_z = img.GetSizeZ();

	 max_x = max_y = max_z = 0;

	 // find the voxels with given label
	 for (size_t i=0; i<img.GetImageSize(); i++)
	 {
		  if (img.GetVoxel(i) == label)
		  {
				x = img.GetX(i);
				y = img.GetY(i);
				z = img.GetZ(i); 
				
				min_x = (min_x > x) ? x : min_x;
				min_y = (min_y > y) ? y : min_y;
				min_z = (min_z > z) ? z : min_z; 
				
				max_x = (max_x < x) ? x : max_x; 
				max_y = (max_y < y) ? y : max_y;
				max_z = (max_z < z) ? z : max_z;
		  }
	 }

	 // form the mask boundaries
	 Vector3d<size_t> label_offset(min_x, min_y, min_z);
	 Vector3d<size_t> label_size(max_x - min_x + 1,
										  max_y - min_y + 1,
										  max_z - min_z + 1);

	 mask.SetResolution(img.GetResolution());
	 mask.SetOffset(img.GetOffset() + 
						 PixelsToMicrons(label_offset, img.GetResolution()));
	 mask.MakeRoom(label_size);
	 mask.Clear();

	 for (x=0; x<label_size.x; x++)
		  for (y=0; y<label_size.y; y++)
				for (z=0; z<label_size.z; z++)
				{
					 if (img.GetVoxel(label_offset.x + x,
											label_offset.y + y,
											label_offset.z + z) == label)
					 {
						  mask.SetVoxel(x,y,z,true);
					 }
				}
}

/*************************************************************************/

template <class T> void StickToImage(Image3d<T> &img,
												 const Image3d<T> &texture,
												 const Image3d<bool> &mask)
{
	 if (mask.GetSize() != texture.GetSize())
		  throw "Error: different size of images!";

	 Offset mask_offset = mask.GetOffset(),
			  img_offset = img.GetOffset(),
			  r_offset = mask_offset - img_offset;

	 Vector3d<size_t> shift = MicronsToPixels(r_offset, img.GetResolution());
	 Vector3d<size_t> sz = mask.GetSize();
	
	 for (size_t i=0; i<mask.GetImageSize(); i++)
	 {
		  if (mask.GetVoxel(i) == true)
		  {
				Vector3d<size_t> pos = mask.GetPos(i);
				T value = texture.GetVoxel(i);

				img.SetVoxel(shift.x + pos.x,
								 shift.y + pos.y,
								 shift.z + pos.z,
								 value);
		  }
	 }
}

/*************************************************************************/

void MergeBinaryImages(Image3d<bool> &img_acc, 
							  const Image3d<bool> &img_added,
							  Vector3d<float> shift)
{
	 Vector3d<size_t> shift_in_pixels;
	 Vector3d<float>  shift_in_pixels_float;

	 shift_in_pixels_float = MicronsToPixels(shift, img_acc.GetResolution());

	 shift_in_pixels.x = (size_t) round(shift_in_pixels_float.x);
	 shift_in_pixels.y = (size_t) round(shift_in_pixels_float.y);
	 shift_in_pixels.z = (size_t) round(shift_in_pixels_float.z);

	 MergeBinaryImages(img_acc, img_added, shift_in_pixels);
}

/*************************************************************************/

void MergeBinaryImages(Image3d<bool> &img_acc, 
							  const Image3d<bool> &img_added,
							  Vector3d<size_t> shift)
{
	 // first, check if the new phantom can be put into the image
	 for (size_t i=0; i<img_added.GetImageSize(); i++)
	 {
		  if (img_added.GetVoxel(i) == true)
		  {
				Vector3d<size_t> pos = img_added.GetPos(i);

				if (img_acc.Include(shift.x + pos.x,
										  shift.y + pos.y,
										  shift.z + pos.z))
				{
					img_acc.SetVoxel(shift.x + pos.x,
										  shift.y + pos.y,
										  shift.z + pos.z,
										  true);
				}

		  }
	 }
}

/*************************************************************************/

inline bool IsInEllipsoid(float x, float y, float z,
		float sx, float sy, float sz,
		float rx, float ry, float rz,
		float rotate_xy, float rotate_xz)
{
	 // first proceed fast rough check
	if (SQR(x-sx) + SQR(y-sy) + SQR(z-sz) > 
		 SQR(std::max(std::max(rx, ry), rz)))
		  return false;

	float x1, x2, x3, y1, y2, y3, z1, z2, z3;

	// shift to center
	x1 = x - sx;
	y1 = y - sy;
	z1 = z - sz;

	// xy-plane rotation
	x2 = float(x1 * cos(DEG_TO_RAD(rotate_xy)) 
				  - y1 * sin(DEG_TO_RAD(rotate_xy)));

	y2 = float(x1 * sin(DEG_TO_RAD(rotate_xy)) 
				  + y1 * cos(DEG_TO_RAD(rotate_xy)));

	z2 = z1;

	// xz-plane rotation
	x3 = float(x2 * cos(DEG_TO_RAD(rotate_xz)) 
				  - z2 * sin(DEG_TO_RAD(rotate_xz)));

	y3 = y2;

	z3 = float(x2 * sin(DEG_TO_RAD(rotate_xz)) 
				  + z2 * cos(DEG_TO_RAD(rotate_xz)));

	return ((SQR(x3/rx) + SQR(y3/ry) + SQR(z3/rz)) <= 1.0f);
}

/*************************************************************************/
inline bool IsInCylinder(float x, float y, float z,
								 float top_x, float top_y, float top_z,
								 float bot_x, float bot_y, float bot_z,
								 float diameter)
{
	 float A = top_x-bot_x,
			 B = top_y-bot_y, 
			 C = top_z-bot_z; // vector elements

	 // check one halfplane (Ax+By+Cz+D>=0). Check whether the point is 
	 // situated above the bottom of the cylinder.
	 float bot_D = -(A*bot_x + B*bot_y + C*bot_z);
	 bool is_above_bottom = (A*x + B*y + C*z + bot_D >= 0);

	 if (!is_above_bottom)
		  return false;

	 // check second halfplane (Ax+By+Cz+D<=0). Chceck whether the point
	 // is situated below the top of the cylinder.
	 float top_D = -(A*top_x + B*top_y + C*top_z);
	 bool is_bellow_top = (A*x + B*y + C*z + top_D <= 0);

	 if (!is_bellow_top)
		  return false;

	 // check tube - create the plane which is perpendicular to the 
	 // axis and passes through the point which is just verified. Find
	 // the intersection of this plane with the axis. The distance between
	 // the point and the intersection is the searched measure.
	 float mid_D = -(A*x + B*y + C*z);
	 // the plane is given by:
	 // Ax+By+Cz+mid_D=0
	 // the axes is given by:
	 // axis_point = (bot_x,bot_y,bot_z) + t*(A,B,C)
	 float t = -(mid_D + A*bot_x + B*bot_y + C*bot_z)/(A*A+B*B+C*C);
	 Vector3d<float> intersec(bot_x+t*A, bot_y+t*B, bot_z+t*C);
	 float dist2 = SQR(intersec.x-x) + SQR(intersec.y-y) + SQR(intersec.z-z);
	 bool is_inside_tube = dist2 <= SQR(diameter/2.0f);

	 if (!is_inside_tube)
		  return false;

	 return true;
}

/*************************************************************************/

template <class VOXEL> void AddEllipsoid(
		Image3d<VOXEL> &img,
		const float centerX,
		const float centerY,
		const float centerZ,
		const float radiusX,
		const float radiusY,
		const float radiusZ,
		const float angle_xy,
		const float angle_xz,
		const VOXEL foreground_intensity)
{
	echo(" ** new ellipsoid ** ");
	echo("foreground color: " << (size_t) foreground_intensity);
	echo("position: (" << centerX << ", " << centerY << ", " <<
			  centerZ << ") microns");
	echo("radius: (" << radiusX << ", " << radiusY << ", " 
		<< radiusZ << ") microns");
	echo("rotation along z-axis(" << angle_xy << ") & y-axis(" << angle_xz 
			  << ") degrees");

	Vector3d<float> res = img.GetResolution().GetRes();
	
	for (size_t i=0; i<img.GetImageSize(); i++)
	{
		if (IsInEllipsoid(
					img.GetX(i)/res.x, img.GetY(i)/res.y, img.GetZ(i)/res.z,
					centerX, centerY, centerZ,
					radiusX, radiusY, radiusZ,
					angle_xy, angle_xz))
		{
			img.SetVoxel(i, foreground_intensity);
		}
	}
}

/*************************************************************************/

template <class VOXEL> void AddCylinder(
		Image3d<VOXEL> &img,
		const float bottomX,
		const float bottomY,
		const float bottomZ,
		const float topX,
		const float topY,
		const float topZ,
		const float diameter,
		const VOXEL foreground_intensity)
{
	echo(" ** new cylinder ** ");
	echo("foreground color: " << (size_t) foreground_intensity);;
	echo("bottom: (" << bottomX << ", " << bottomY << ", " <<
			  bottomZ << ") microns");
	echo("top: (" << topX << ", " << topY << ", " << topZ << ") microns");

	Vector3d<float> res = img.GetResolution().GetRes();
	
	for (size_t i=0; i<img.GetImageSize(); i++)
	{
		if (IsInCylinder(
					img.GetX(i)/res.x, 
					img.GetY(i)/res.y, 
					img.GetZ(i)/res.z,
					bottomX, 
					bottomY, 
					bottomZ,
					topX, 
					topY, 
					topZ,
					diameter))
		{
			img.SetVoxel(i, foreground_intensity);
		}
	}
}

/*************************************************************************/

template <class VOXEL> void AddSnake(
		Image3d<VOXEL> &img,
		std::vector<Vector3d<float>/**/> &nodes,
		const float diameter,
		const VOXEL foreground_intensity)
{
	echo(" ** new snake ** ");
	echo("foreground color: " << (size_t) foreground_intensity);
	echo("bottom: (" << nodes[0].x << ", " << nodes[0].y << ", " <<
			  nodes[0].z << ") microns");

#ifdef DEBUG	
	size_t last = nodes.size()-1;
	echo("top: (" << nodes[last].x << ", " << nodes[last].y << ", " 
		<< nodes[last].z << ") microns"); 
#endif

	// first, generate snake axis using splines driven by nodes
	// make a new image and put the snake axis into this image. After that,
	// call distance transform function in order to get the set of all the 
	// pixel which are in the given distance from this axes. The snake is then
	// simple obtained as a thresholding.

	// allocate the memory
	size_t sz = nodes.size();
   double *steps = new double[sz],
			*x_array = new double[sz],
			*y_array = new double[sz],
			*z_array = new double[sz];

	// prepare the data
	for (size_t i=0; i<sz; i++)
	{
		 steps[i] = i;
		 x_array[i] = nodes[i].x;
		 y_array[i] = nodes[i].y;
		 z_array[i] = nodes[i].z;
	}

	// initialize gsl spline engine
	gsl_interp_accel *x_accel_ptr = gsl_interp_accel_alloc(),
						  *y_accel_ptr = gsl_interp_accel_alloc(),
						  *z_accel_ptr = gsl_interp_accel_alloc();

	gsl_spline *x_spline_ptr = gsl_spline_alloc(gsl_interp_cspline, sz);
	gsl_spline *y_spline_ptr = gsl_spline_alloc(gsl_interp_cspline, sz);
	gsl_spline *z_spline_ptr = gsl_spline_alloc(gsl_interp_cspline, sz);

	// find the spline curve component along each axis separately
	/*float dx = (xs[sz] - xs[1])/(param[sz] - param[1]);
	spline(param, xs, sz, dx, dx, x2s);
	
	float dy = (ys[sz] - ys[1])/(param[sz] - param[1]);
	spline(param, ys, sz, dy, dy, y2s);

	float dz = (zs[sz] - zs[1])/(param[sz] - param[1]);
	spline(param, zs, sz, dz, dz, z2s);
	*/

	echo("initialize GSL spline");
	gsl_spline_init(x_spline_ptr, steps, x_array, sz);
	gsl_spline_init(y_spline_ptr, steps, y_array, sz);
	gsl_spline_init(z_spline_ptr, steps, z_array, sz);

	// allocate the memory block for a new (temporary) image
	Image3d<float> dist_image;
	dist_image.MakeRoom(img.GetSize());
	dist_image.SetResolution(img.GetResolution());
	dist_image.SetOffset(img.GetOffset());

	// draw the spline axis into the image
	Vector3d<float> res = img.GetResolution().GetRes();
	for (double step=steps[0]; step <= steps[sz-1]; step+=0.01)
	{
		 double x,y,z;

		 // smazat: echo("eval for " << step);
		 x = gsl_spline_eval (x_spline_ptr, step, x_accel_ptr);
		 y = gsl_spline_eval (y_spline_ptr, step, y_accel_ptr);
		 z = gsl_spline_eval (z_spline_ptr, step, z_accel_ptr);
		 
		 size_t ix = (size_t) round(x*res.x);
		 size_t iy = (size_t) round(y*res.y);
		 size_t iz = (size_t) round(z*res.z);

		 if (dist_image.Include(ix, iy, iz))
		 	dist_image.SetVoxel(ix, iy, iz, 1);
	}

	// echo("distance transform");
	
	// call distance transform
	EDM(dist_image, 0, 0.0f, true);

	float radius = diameter/2;

	for (size_t i=0; i<dist_image.GetImageSize(); i++)
	{
		 if (dist_image.GetVoxel(i) <= radius)
		 {
			  img.SetVoxel(i, foreground_intensity);
		 }
	}

	// release occupied memory
	delete [] steps;
	delete [] x_array;
	delete [] y_array;
	delete [] z_array;

	gsl_spline_free(x_spline_ptr);
	gsl_spline_free(y_spline_ptr);
	gsl_spline_free(z_spline_ptr);

	gsl_interp_accel_free(x_accel_ptr);
	gsl_interp_accel_free(y_accel_ptr);
	gsl_interp_accel_free(z_accel_ptr);
}

/*************************************************************************/
/* explicit instantiations */
template  void AddEllipsoid(
		Image3d<bool> &img,
		const float centerX,
		const float centerY,
		const float centerZ,
		const float radiusX,
		const float radiusY,
		const float radiusZ,
		const float angle_xy,
		const float angle_xz,
		const bool foreground_intensity);

template  void AddEllipsoid(
		Image3d<unsigned> &img,
		const float centerX,
		const float centerY,
		const float centerZ,
		const float radiusX,
		const float radiusY,
		const float radiusZ,
		const float angle_xy,
		const float angle_xz,
		const unsigned foreground_intensity);

template  void AddEllipsoid(
		Image3d<GRAY8> &img,
		const float centerX,
		const float centerY,
		const float centerZ,
		const float radiusX,
		const float radiusY,
		const float radiusZ,
		const float angle_xy,
		const float angle_xz,
		const GRAY8 foreground_intensity);

template  void AddEllipsoid(
		Image3d<GRAY16> &img,
		const float centerX,
		const float centerY,
		const float centerZ,
		const float radiusX,
		const float radiusY,
		const float radiusZ,
		const float angle_xy,
		const float angle_xz,
		const GRAY16 foreground_intensity);

template  void AddEllipsoid(
		Image3d<float> &img,
		const float centerX,
		const float centerY,
		const float centerZ,
		const float radiusX,
		const float radiusY,
		const float radiusZ,
		const float angle_xy,
		const float angle_xz,
		const float foreground_intensity);

template  void AddCylinder(
		Image3d<bool> &img,
		const float bottomX,
		const float bottomY,
		const float bottomZ,
		const float topX,
		const float topY,
		const float topZ,
		const float diameter,
		const bool foreground_intensity);

template  void AddCylinder(
		Image3d<GRAY8> &img,
		const float bottomX,
		const float bottomY,
		const float bottomZ,
		const float topX,
		const float topY,
		const float topZ,
		const float diameter,
		const GRAY8 foreground_intensity);

template  void AddCylinder(
		Image3d<GRAY16> &img,
		const float bottomX,
		const float bottomY,
		const float bottomZ,
		const float topX,
		const float topY,
		const float topZ,
		const float diameter,
		const GRAY16 foreground_intensity);

template void AddSnake(
		Image3d<bool> &img,
		std::vector<Vector3d<float>/**/> &nodes,
		const float diameter,
		const bool foreground_intensity);

template void AddSnake(
		Image3d<GRAY8> &img,
		std::vector<Vector3d<float>/**/> &nodes,
		const float diameter,
		const GRAY8 foreground_intensity);

template void AddSnake(
		Image3d<GRAY16> &img,
		std::vector<Vector3d<float>/**/> &nodes,
		const float diameter,
		const GRAY16 foreground_intensity);

template void AddSnake(
		Image3d<unsigned> &img,
		std::vector<Vector3d<float>/**/> &nodes,
		const float diameter,
		const unsigned foreground_intensity);

template void StickToImage(Image3d<GRAY8> &, const Image3d<GRAY8> &, 
									const Image3d<bool> &);
template void StickToImage(Image3d<GRAY16> &, const Image3d<GRAY16> &, 
									const Image3d<bool> &);
template void StickToImage(Image3d<float> &, const Image3d<float> &, 
									const Image3d<bool> &);
template void StickToImage(Image3d<double> &, const Image3d<double> &, 
									const Image3d<bool> &);

template void GetLabelMask(const i3d::Image3d<size_t> &, size_t, 
									i3d::Image3d<bool> &);

template void GetLabelMask(const i3d::Image3d<GRAY16> &, GRAY16, 
									i3d::Image3d<bool> &);

template bool AddNewLabel(i3d::Image3d<GRAY16> &img,
								  const i3d::Image3d<bool>& mask,
								  bool ignore_overlaps);

template bool AddNewLabel(i3d::Image3d<size_t> &img,
								  const i3d::Image3d<bool>& mask,
								  bool ignore_overlaps);

