/*---------------------------------------------------------------------

collisions.cc

This file is part of CytoPacq

Copyright (C) 2007-2013 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Author: David Svoboda <svoboda@fi.muni.cz>

Description: A simple (correlation based) solution for 
   'object collision problem'.

-------------------------------------------------------------------------*/

#include <float.h>
#include <i3d/DistanceTransform.h>
#include <i3d/morphology.h>
#include <i3d/se.h>
#include <i3d/convolution.h>

#include "sim_macros.h"
#include "collisions.h"

#define sqr(x) ((x)*(x))

using namespace i3d;
using namespace std;

#define SMALL_VALUE 0.5f
#define SCALING_FACTOR 100
#define TOLERANCE_IN_MICRONS 0.4

/*
//---------------------------------------------------------------------------
// This function takes the first argument as an input value a normalize
// it respecting the range defined by the second and the third argument.
// The results is a value between 0 and 1 which is further submitted to
// skewed exponential function (optionally multiplied to get the values
// in the range [0; SCALING_FACTOR]
//---------------------------------------------------------------------------
size_t normalize_and_scale(float value,
									float l_range,
									float r_range,
									float weight,
									float scaling = SCALING_FACTOR)
{
	 const float skew = 30.0f; // skewness of exponential function

	 float factor  = (value - l_range) / (r_range - l_range);
	 
	 return scaling * weight * exp((-1.0f) * skew * weight * factor);
}
*/

//---------------------------------------------------------------------------
template <class T>
bool GetLastNonZeroPosition(const Image3d<T> &img, Vector3d<size_t> &v)
{
	 for (size_t i = img.GetImageSize()-1; i>=0; i--)
	 {
		  if (img.GetVoxel(i) != T(0))
		  {
				v = img.GetPos(i);
				return true;
		  }
	 }

	 return false;
}

//---------------------------------------------------------------------------
// This function get two inputs: the mask of available area (0 - free,
// 1 - occupied) and the mask of a new object (1 - foreground, 0 - background)
// which we want to put into any free position of given available area.
// These two images are submitted to the correlation function. All the possible
// positions, where the object might be placed, correspond to zero correlation
// coefficients.
// 
// In order to guarantee one pixel gap between the object and the surrounding
// objects the object is apriori submitted to one round of dilation.
//
// The positions are weighted with the distance from the already 
// generated objects.
//
// This way, we generate the clustering effect of objects (cells).
//
// The detected position is the position of the center of the image which 
// contains the observed object.
//---------------------------------------------------------------------------

Vector3d<float> *find_position_in_cluster(const Image3d<bool> &object, 
														const Image3d<bool> &space,
														float cluster_effect,
														bool influence_of_gravity)
{
	 Image3d<bool> object_mirror(object), space_mirror;
	 Image3d<float> correlation_results;

	 echo("dilation of space image");

	 // dilate the space to guarantee there will be a gap between
	 // the generated objects
	 Dilation(space, space_mirror, nb3D_26);

	 echo("drawing border line");

	 // draw a border line in the space to prevent the possible object
	 // positions from being located at the border. 
	 for (size_t i=0; i<space_mirror.GetImageSize(); i++)
	 {
		  Vector3d<size_t> coordinates = space_mirror.GetPos(i);

		  if (space_mirror.OnBorder(coordinates))
		  {
				space_mirror.SetVoxel(i, true);
		  }
	 }

    echo("performing convolution");

	 // apply convolution; as we flip the image in advance we apply
	 // the correlation instead of convolution
	 object_mirror.Flip(true, true, true);
	 Convolution<double>(space_mirror, 
								object_mirror, 
								correlation_results, 
								false);

	 object_mirror.DisposeData();
	 space_mirror.DisposeData();

	 // Find the weights by distance transform:
	 Image3d<bool> free_space;
	 free_space.CopyMetaData(correlation_results);

	 echo("thresholding convolution result");

	 // Mask the results of distance transform with the correlation results
	 for (size_t i=0; i<correlation_results.GetImageSize(); i++)
	 {
		  // remove the positions, where the correlation results
		  // give some non-zero value
		  if (correlation_results.GetVoxel(i) < SMALL_VALUE)
				free_space.SetVoxel(i, true);
		  else
				free_space.SetVoxel(i, false);
	 }

	 correlation_results.DisposeData();

	 // if required impose the gravity to the objects
	 if (influence_of_gravity)
	 {
		  Vector3d<float> res = free_space.GetResolution().GetRes();
		  size_t tolerance_in_pixels = 
					 (size_t) floorf(TOLERANCE_IN_MICRONS * res.z);

		  Vector3d<size_t> pos;

		  echo("tolerance in pixels: " << tolerance_in_pixels);

		  if (!GetLastNonZeroPosition(free_space, pos))
		  {
				throw ERROR_DETAILS("No more space for new object");
		  }

		  echo("last non zero position: " << pos);
		  size_t spread = 
			 (pos.z >= tolerance_in_pixels) ? (pos.z - tolerance_in_pixels) : 0;

		  for (size_t i=0; i<free_space.GetImageSize(); i++)
		  {
				if (free_space.GetZ(i) < spread)
					 free_space.SetVoxel(i, false);
		  }
	 }

	 // Respect the cluster effect when generating new positions.
	 bool cluster = (((float)rand())/RAND_MAX) < (cluster_effect/100.0f);

//	 free_space.SaveImage("gravitace.ics");

	 // Apply cluster effect only if there exists at least one object. Otherwise,
	 // it has no sense.
	 if (cluster && (space.GetMaxValue() == true))
	 {
		  echo("cluster effect: yes");
		  Image3d<float> distance_map;

		  distance_map.CopyMetaData(space);

		  for (size_t i=0; i<distance_map.GetImageSize(); i++)
		  {
				if (space.GetVoxel(i) == true)
					 distance_map.SetVoxel(i, 1.0f);
				else
					 distance_map.SetVoxel(i, 0.0f);
		  }

		  EDM(distance_map, 0, 0.0f, false);

//		  distance_map.SaveImage("mapa.ics");

		  for (size_t i=0; i<distance_map.GetImageSize(); i++)
		  {
				if (free_space.GetVoxel(i) == false)
					 distance_map.SetVoxel(i, FLT_MAX);
		  }

		  float min_distance = distance_map.GetMinValue();

		  for (size_t i=0; i<distance_map.GetImageSize(); i++)
		  {
				if (free_space.GetVoxel(i) == true)
				{
					 float dist_value = distance_map.GetVoxel(i);

					 // je vzdalenost v pixelech nebo mikronech???
					 if ((dist_value < min_distance) ||
						  (dist_value > min_distance + TOLERANCE_IN_MICRONS))
					 {
						  free_space.SetVoxel(i, false);
					 }
				}
		  }

		  //free_space.SaveImage("cluster.ics");
	 }
#ifdef DEBUG
	 free_space.SaveImage("debug-free-space.ics");
#endif

	 // check the number of free positions
	 size_t num_of_positions = 0;

	 for (size_t i=0; i<free_space.GetImageSize(); i++)
	 {
		  if (free_space.GetVoxel(i) == true)
				num_of_positions++;
	 }

    echo("sum of free positions: " << num_of_positions);

	 // If no suitable position was found, leave the function with no success.
	 if (num_of_positions == 0)
	 {
		  return NULL;
	 }

	 // Find the suitable position 
	 size_t barrier = (size_t) ((((float)rand()) / RAND_MAX)*num_of_positions);
	 size_t index = 0, shift = 0;

	 while (shift <= barrier)
	 {
		  shift += (free_space.GetVoxel(index)) ? 1 : 0;
		  index++;
	 }

	 Vector3d<float> *position = new Vector3d<float>;

	 *position = PixelsToMicrons(free_space.GetPos(index), 
										  free_space.GetResolution());

	 echo("searching finished");

	 return position;
}

//---------------------------------------------------------------------------
Vector3d<float> *find_position_uniformly(const Image3d<bool> &object,
													  const Image3d<bool> &space,
													  bool influence_of_gravity)
{
	 return find_position_in_cluster(object, space, 0.0f, influence_of_gravity);
}

//--------------------------------------------------------------------------

Vector3d<float> 
*find_position_gaussian_weighted(const Image3d<bool> &object,
											const Image3d<bool> &space,
											Vector3d<float> gauss_mean,
											Vector3d<float> gauss_dev,
											bool influence_of_gravity)
{
	 Image3d<bool> object_mirror (object), space_mirror;
	 Image3d<float> correlation_results;

	 // dilate the space to guarantee there will be a gap between
	 // the generated objects
	 Dilation(space, space_mirror, nb3D_26);

	 // draw a border line in the space to prevent the possible object
	 // positions from being found near the border. 
	 for (size_t i=0; i<space_mirror.GetImageSize(); i++)
	 {
		  Vector3d<size_t> coordinates = space_mirror.GetPos(i);

		  if (space_mirror.OnBorder(coordinates))
		  {
				space_mirror.SetVoxel(i, true);
		  }
	 }

	 // apply convolution; as we flipped the image in advance we apply
	 // the correlation instead of convolution
	 object_mirror.Flip(true, true, true);
	 Convolution<double>(space_mirror, 
								object_mirror, 
								correlation_results, 
								false);

	 object_mirror.DisposeData();
	 space_mirror.DisposeData();

	 // Find the weights by distance transform:
	 Image3d<bool> free_space;
	 free_space.CopyMetaData(correlation_results);

	 echo("thresholding convolution result");

	 // Mask the results of distance transform with the correlation results
	 for (size_t i=0; i<correlation_results.GetImageSize(); i++)
	 {
		  // remove the positions, where the correlation results
		  // give some non-zero value
		  if (correlation_results.GetVoxel(i) < SMALL_VALUE)
				free_space.SetVoxel(i, true);
		  else
				free_space.SetVoxel(i, false);
	 }

	 correlation_results.DisposeData();

	 // if required impose the gravity to the objects
	 if (influence_of_gravity)
	 {
		  Vector3d<float> res = free_space.GetResolution().GetRes();
		  size_t tolerance_in_pixels = 
					 (size_t) floorf(TOLERANCE_IN_MICRONS * res.z);

		  Vector3d<size_t> pos;

		  echo("tolerance in pixels: " << tolerance_in_pixels);

		  if (!GetLastNonZeroPosition(free_space, pos))
		  {
				throw ERROR_DETAILS("No more space for new object");
		  }

		  echo("last non zero position: " << pos);
		  size_t spread = 
			 (pos.z >= tolerance_in_pixels) ? (pos.z - tolerance_in_pixels) : 0;

		  for (size_t i=0; i<free_space.GetImageSize(); i++)
		  {
				if (free_space.GetZ(i) < spread)
					 free_space.SetVoxel(i, false);
		  }
	 }

	 // generate 3D Gaussian function - we will generate pixel positions
	 // each containing the probability function
	 // following the normal (Gaussian) distribution.
	 float gauss_x, gauss_y, gauss_z;
	 Vector3d<float> res = space.GetResolution().GetRes();
	 Image3d<GRAY16> gauss_img;
	 gauss_img.MakeRoom(space.GetSize());

	 // generate normalized Gaussian hat over the whole 3D image
	 for (size_t x=0; x<gauss_img.GetSizeX(); x++)
		  for (size_t y=0; y<gauss_img.GetSizeY(); y++)
				for (size_t z=0; z<gauss_img.GetSizeZ(); z++)
				{
	 				 Vector3d<float> real_pos(x/res.x, y/res.y, z/res.z);

					 gauss_x = exp((-0.5)*sqr(real_pos.x - gauss_mean.x) / 
										sqr(gauss_dev.x));
					 gauss_y = exp((-0.5)*sqr(real_pos.y - gauss_mean.y) / 
										sqr(gauss_dev.y));
					 gauss_z = exp((-0.5)*sqr(real_pos.z - gauss_mean.z) / 
										sqr(gauss_dev.z));
					 float value = SCALING_FACTOR * gauss_x * gauss_y * gauss_z;

					 gauss_img.SetVoxel(x,y,z,(GRAY16) floorf(value));
				}

	 // Sum up the Gaussian (hat) function over the whole image (within
	 // free area only).
	 size_t sum_of_weights = 0;

	 for (size_t i=0; i<free_space.GetImageSize(); i++)
	 {
		  if (free_space.GetVoxel(i))
		  {
		  		sum_of_weights += gauss_img.GetVoxel(i);
		  }
	 }

    echo("sum of weights: " << sum_of_weights);

	 // If no suitable position was found, leave the function with no success.
	 if (sum_of_weights == 0)
	 {
		  return NULL;
	 }
	 
	 // find random position within the area of available positions
	 size_t barrier = (size_t) ((((float)rand()) / RAND_MAX)*sum_of_weights);
	 size_t index = 0, shift = 0;

	 while (shift <= barrier)
	 {
		  if (free_space.GetVoxel(index))
		  {
				shift += gauss_img.GetVoxel(index);
		  }
		  index++;
	 }

	 Vector3d<float> *position = new Vector3d<float>;

	 *position = PixelsToMicrons(space.GetPos(index), space.GetResolution());

	 echo("searching finished");

	 return position;
}

