/*---------------------------------------------------------------------

diagram.cc

This file is part of CytoPacq

Copyright (C) 2009-2013 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.


Authors: 
- David Svoboda <svoboda@fi.muni.cz> May 2010
- Ondrej Homola <ondra.homola@gmail.com> June 2009


Description: The implementation of some geometric primitives in 3D space
             (plane, polyhedron, voronoi diagram). The voronoi diagram
				 is created using the library - LIBQHULL (www.qhull.org).

				 The definition of these classes and their methods:
				 - Plane
				 - Polyhedron
				 - VoronoiDiagram

				 The definition of these function:
				 - get_area(...)
				 - get_volume(...)

-------------------------------------------------------------------------*/

#include <stdio.h>
#include <fstream>

extern "C"
{
#include <qhull_a.h>
}

#include "diagram.h"
#include "sim_macros.h"

using namespace std;

//----------------------------------------------------------------------

#ifdef _MSC_VER
	#define GET_UNIQUE_FNAME _tempnam
#else
	#define GET_UNIQUE_FNAME tempnam
#endif

//----------------------------------------------------------------------
/**
 * His function defines the area of triangle defined by the points
 * A,B, and C. The lower the value is returned the more acute the angle
 * between the points A,B, and C is.
 */
//----------------------------------------------------------------------
REAL get_area(const Position& A, const Position& B, const Position& C)
{
	 // Create two vectors (A->B and A->C) 
	 Vector u = B - A, v = C - A;
	 
	 // Make the cross product of these two vectors
	 Vector perpendicular = Cross(u,v);

	 // Norm of cross product defines the area of parallelogram defined
	 // we the vectors 'u' and 'v'. As we are interested in the area of 
	 // triangle, we will divide the solution by 2.
	 return Norm(perpendicular)/2;
}

//----------------------------------------------------------------------
/**
 * His function defines the volume of tetrahedron defined by the points
 * A,B,C, and D. 
 */
//----------------------------------------------------------------------
REAL get_volume(const Position& A, const Position& B, 
					 const Position& C, const Position& D)
{
	 // Create three vectors (A->B, A->C, A->D) 
	 Vector u = B - A, v = C - A, w = D - A;

	 // Volume = |u . (v x w)|
	 
	 // Make the cross product of 'v' & 'w' followed by dot product with 'u':
	 Vector result = Dot(u,Cross(v,w));

	 // Norm of cross product defines the area of parallelogram defined
	 // we the vectors 'u' and 'v'. As we are interested in the area of 
	 // triangle, we will divide the solution by 2.
	 return Norm(result)/6;
}

//----------------------------------------------------------------------
//                          Plane methods
//----------------------------------------------------------------------

Plane::Plane(const Position &p1, const Position &p2, const Position &p3)
{
	 // Take the input arguments (three points) and define the plane
	 // in the standard way (normal vector & offset)
	 Vector u = p2 - p1,
			  v = p3 - p1;
	 
	 normal_vector = Cross(u,v);

	 offset = -Dot(normal_vector, p1);
}

//----------------------------------------------------------------------

Plane::Plane(const Vector &v, REAL _offset):
		  normal_vector(v),offset(_offset)
{
}

//----------------------------------------------------------------------

bool Plane::has_on_negative_side(const Position &_point) const
{
	 // The point '_point' is located on the left side, if the 
	 // plane equation Ax+By+Cz+D applied on this point returns
	 // the number lower or equal to zero.
	 return ((Dot(normal_vector,_point) + offset) <= 0);
}

//----------------------------------------------------------------------

bool Plane::has_on_positive_side(const Position &_point) const
{
	 // The point '_point' is located on the right side, if the 
	 // plane equation Ax+By+Cz+D applied on this point returns
	 // the number greater than zero.
	 return ((Dot(normal_vector,_point) + offset) > 0);
}

//----------------------------------------------------------------------

Plane Plane::opposite() const
{
	 // Let us flip the plane. The plane defined is left unchanged. The
	 // new plane with an opposite orientation is created and returned.
	 return Plane(-normal_vector, -offset);
}

//----------------------------------------------------------------------
//                    Polyhedron methods
//----------------------------------------------------------------------

Polyhedron::Polyhedron()
{
	 // initially the bounding box of a polyhedron is not defined.
	 bounding_points_found = false;

	 // initially the volume is set to zero
	 volume = 0;
}

//----------------------------------------------------------------------

Polyhedron::Polyhedron(const std::vector<Plane> &_planes, 
							  const std::vector<Position> &_vertices)
{
	// create the polyhedron and fill its internal variables
	planes = _planes;
	vertices = _vertices;
	volume = 0;

	// if both the input argument correspond to empty vectors, it has
	// no sense to create such a polyhedron
	if (planes.empty() || vertices.empty())
	{
		 throw "Polyhedron::Polyhedron: empty set.";
	}

	// establish bound box of this polyhedron
	FindBoundingPoints();
}

//----------------------------------------------------------------------

void Polyhedron::FindBoundingPoints()
{
	// Track all the polyhedron's vertices and
	// find top and low bounding point:
	top_bounding_point = low_bounding_point = vertices[0];

	for (size_t i = 1; i < vertices.size(); i++)
	{
		 top_bounding_point = min(top_bounding_point, vertices[i]);
		 low_bounding_point = max(low_bounding_point, vertices[i]);
	}

	bounding_points_found = true;
}

//----------------------------------------------------------------------

void Polyhedron::AddPlane(const Plane& _plane)
{
	 // Add new plane the to polyhedron (apply new constraint)
	 planes.push_back(_plane);
}

//----------------------------------------------------------------------

void Polyhedron::IncrementVolume(double amount)
{
	 volume += amount;
}

//----------------------------------------------------------------------

Position Polyhedron::GetTopBoundingPoint() const
{
	 // if the bounding box is not defined we cannot return the top point
	 if (bounding_points_found == false)
		  throw "Polyhedron::GetTopBoundingPoint(): bounding points not set";

	 return top_bounding_point;
}

//----------------------------------------------------------------------

Position Polyhedron::GetLowBoundingPoint() const
{
	 // if the bounding box is not defined we cannot return the bottom point
	 if (bounding_points_found == false)
		  throw "Polyhedron::GetLowBoundingPoint(): bounding points not set";

	 return low_bounding_point;
}

//----------------------------------------------------------------------

void Polyhedron::AddVertex(const Position& _vertex)
{
	 // add a new vertex to the polyhedron
	 vertices.push_back(_vertex);

	 // adding a new point makes the current bounding box invalid
	 bounding_points_found = false;
}

//----------------------------------------------------------------------

bool Polyhedron::contains(const Position &p) const
{
	 // Track all the polyhedron's planes and check if these have the same
	 // orientation respecting the given point 'p'. If so and the orientation
	 // is positive, the point is considered to be located inside the 
	 // polyhedron.
	for (size_t i = 0; i < planes.size(); i++)
	{	
		if (planes[i].has_on_negative_side(p))
		{			
			return false;
		}
	}
	return true;
}

//----------------------------------------------------------------------
//                 Vornoi Diagram methods
//----------------------------------------------------------------------

void VoronoiDiagram::Clear()
{
	 // Delete all the voronoi regions, if there are some already defined.
	 regions.clear();
}

//----------------------------------------------------------------------

VoronoiDiagram::VoronoiDiagram()
{
	 // create unique names for QHULL i/o file buffers
	 qhull_input_fname = GET_UNIQUE_FNAME(NULL,"qhull");
	 qhull_output_fname = GET_UNIQUE_FNAME(NULL,"qhull");
	 qhull_log_fname = GET_UNIQUE_FNAME(NULL,"qhull");

	 // debug info - type out the name of the buffers
	 echo("points: " << qhull_input_fname);
	 echo("voronoi diagram: " << qhull_output_fname);
	 echo("logfile: " << qhull_log_fname);

	 Clear();
}

//----------------------------------------------------------------------

VoronoiDiagram::~VoronoiDiagram()
{
	 // Delete (from file system) the QHULL i/o file buffers.
	 remove(qhull_input_fname.c_str());
	 remove(qhull_output_fname.c_str());
	 remove(qhull_log_fname.c_str());
}

//----------------------------------------------------------------------

void VoronoiDiagram::PrepareQhullInput(const std::vector<Position> &cell_points,
										  const std::vector<Position> &control_points)
{
	 /*
	  * Format of qhull input file:
	  *
	  * <dimension>
	  * <number of triangulated points>
	  * <point1 cartesian coordinates>
	  * <point2 cartesian coordinates>
	  * <point3 cartesian coordinates>
	  * ...
	  * <pointN cartesian coordinates>
	  */

	 // create an output stream
	 ofstream ofs(qhull_input_fname.c_str());

	 // If the output stream was not created correctly, do not proceed.
	 if (!ofs.good())
	 {
		  throw "VoronoiDiagram::PrepareQhullInput(): Cannot open file.";
	 }

	 // define the dimension
	 int dim = 3;
	 ofs << dim << endl;

	 // set number of points
	 ofs << cell_points.size() + control_points.size() << endl;

	 // define the delaunay triangulation vertices
	 // and put them to the output stream
	 for (size_t i=0; i<cell_points.size(); i++)
	 {
		  ofs << cell_points[i].x << " " <<
					 cell_points[i].y << " " <<
					 cell_points[i].z << endl;
	 }

	 for (size_t i=0; i<control_points.size(); i++)
	 {
		  ofs << control_points[i].x << " " <<
					 control_points[i].y << " " <<
					 control_points[i].z << endl;
	 }

	 // close the output stream
	 ofs.close();
}

//----------------------------------------------------------------------

void VoronoiDiagram::CallQhull()
{
	 // start computing Voronoi diagram
	 // and prepare the QHULL-specifiac variables
	 boolT ismalloc;
	 int dim, curlong, totlong, numpoints;

	 // prepare QHULL i/o buffers
	 FILE *qhull_input = fopen(qhull_input_fname.c_str(),"r"),
			*qhull_output = fopen(qhull_output_fname.c_str(),"w"),
			*qhull_log = fopen(qhull_log_fname.c_str(),"w");

	 // if the opening of some buffer failed, throw an exception
	 if ((qhull_input == NULL) ||
		  (qhull_output == NULL) ||
		  (qhull_log == NULL))
	 {
		  throw "VoronoiDiagram::CallQhull(): Cannot open file.";
	 }

	 // init QHULL
	 qh_meminit(qhull_log);
	 qh_initqhull_start(qhull_input, qhull_output, qhull_log);

	 // set the options (read the source codes in QHULL, there is no
	 // appropriate documentation for QHULL) for QHULL to generate
	 // voronoi diagram
	 qh_option("voronoi _bbound-last _coplanar-keep offFile", NULL, NULL);
    qh_appendprint(qh_PRINToff);
	 qh_appendprint(qh_PRINTvertices);

	 qh DELAUNAY = True;
	 qh VORONOI = True;
	 qh SCALElast = True;

	 // load the input information
	 coordT *points = qh_readpoints(&numpoints, &dim, &ismalloc);
	 qh_init_B(points, numpoints, dim, ismalloc);

	 // create the voronoi diagram
	 qh_qhull();

	 // put the results into output buffers (files)
	 qh_check_output();
	 qh_produce_output();

	 // release the memory
	 qh_freeqhull( False);
	 qh_memfreeshort(&curlong, &totlong);

	 // close the buffers (files)
	 fclose(qhull_input);
	 fclose(qhull_output);
	 fclose(qhull_log);
}

//----------------------------------------------------------------------

void VoronoiDiagram::ParseQhullOutput(double diagram_skew_x,
												  double diagram_skew_y,
												  double diagram_skew_angle)
{
	 // Open the file in which the QHULL solution is stored.
	 ifstream ifs(qhull_output_fname.c_str());

	 if (!ifs.good())
	 {
		  throw "VoronoiDiagram::ParseQhullOutput(): Cannot open file.";
	 }

	 /* Read the information concerning the Voronoi region following
	  * this syntax (keep in mind that this syntex is driven by the
	  * QHULL options, i.e. this is not a general output format!
	  *
	  * % begin of file
	  * <dimension>
	  * <number of vertices> <number of regions> <number of ridges>
	  * % list of vertices
	  * % the zero-th vertex is always placed in infinity
	  * <vertex1 cartesian coordinates>
	  * <vertex2 cartesian coordinates>
	  * ...
	  * <vertexN cartesian coordinates>
	  * % list of voronoi regions (VR)
	  * <number of vertices of VR1> <Vertex1Index> <Vertex2Index> ...
	  * <number of vertices of VR2> <Vertex1Index> <Vertex2Index> ...
	  * ...
	  * <number of vertices of VRM> <Vertex1Index> <Vertex2Index> ...
	  * % list of facets (planes)
	  * <number of planes>
	  * <num of items of Plane1> <LeftVRIndex> <RightVRIndex> <Vertex1Index>
	  * 		<Vertex2Index> <Vertex3Index> ...
	  * <num of items of Plane2> <LeftVRIndex> <RightVRIndex> <Vertex1Index>
	  * 		<Vertex2Index> <Vertex3Index> ...
	  * ...
	  * <num of items of PlaneK> <LeftVRIndex> <RightVRIndex> <Vertex1Index>
	  * 		<Vertex2Index> <Vertex3Index> ...
	  * % EOF
	  */
	 int dim;
	 ifs >> dim;

	 size_t num_of_vertices, num_of_regions, num_of_ridges;

	 ifs >> num_of_vertices >> num_of_regions >> num_of_ridges;

	 // Read in the vertices of Voronoi diagram and simultaneously apply
	 // affine transform (skew) on each vertex. 
	 std::valarray<Position> voronoi_vertices;
	 voronoi_vertices.resize(num_of_vertices);

	 // debug info
	 echo("diagram_skew_angle: " << diagram_skew_angle);
	 echo("diagram_skew_x: " << diagram_skew_x);
	 echo("diagram_skew_y: " << diagram_skew_y);
	 echo("skew x: " << diagram_skew_x * cos(DEG_TO_RAD(diagram_skew_angle)));
	 echo("skew y: " << diagram_skew_y * sin(DEG_TO_RAD(diagram_skew_angle)));

	 // It is possible that we want to skew the output Voronoi diagram.
	 // In that case we will apply the affine transform (skew) to the 
	 // coordinates of each vertex in voronoi diagram.
	 for (size_t i=0; i<num_of_vertices; i++)
	 {
		  REAL x,y,z;
		  ifs >> x >> y >> z;

		  // perform the affine transformation (skew)
		  REAL aft_x, aft_y, aft_z;
		  aft_x = x + z * diagram_skew_x * cos(DEG_TO_RAD(diagram_skew_angle));
		  aft_y = y + z * diagram_skew_y * sin(DEG_TO_RAD(diagram_skew_angle));
		  aft_z = z;

		  voronoi_vertices[i] = Position(aft_x, aft_y, aft_z);
	 } // loop (vertices)

	 // read in the Voronoi regions
	 INDEX vertex_index;

	 for (size_t i=0; i<num_of_regions; i++)
	 {
		  size_t amount; // number of vertices belonging to the current region
		  ifs >> amount;

		  for (size_t j=0; j<amount; j++)
		  {
				ifs >> vertex_index;
				regions[i].AddVertex(voronoi_vertices[vertex_index]);
		  }

		  regions[i].FindBoundingPoints();

		  // In the previous loop we skewed the vertices of Voronoi diagram.
		  // Do not forget to apply the same transform to Voronoi points,
		  // i.e. to the triangulated points in Delaunay triangulation.
		  Position c = regions[i].GetCenter();
		  Position aft_c;

		  aft_c.x = c.x + 
					 c.z * diagram_skew_x * cos(DEG_TO_RAD(diagram_skew_angle));
		  aft_c.y = c.y + 
					 c.z * diagram_skew_y * sin(DEG_TO_RAD(diagram_skew_angle));
		  aft_c.z = c.z;

		  regions[i].SetCenter(aft_c);
	 } // loop (regions)

	 // read in the Voronoi planes
	 size_t num_of_planes;
	 ifs >> num_of_planes;

	 for (size_t i=0; i<num_of_planes; i++)
	 {
		  size_t amount; // number of indices
		  ifs >> amount;

		  // read the index of the first adjacent voronoi region
		  INDEX left_voronoi_region;
		  ifs >> left_voronoi_region; 
		  
		  // read the index of the second adjacent voronoi region
		  INDEX right_voronoi_region;
		  ifs >> right_voronoi_region;

		  // lower the variable amount as we have just read two indices
		  amount = amount - 2;

		  if (amount < 3)
				throw "VoronoiDiagram::ParseQhullOutput(): "
						  " incorrectly defined plane";

		  INDEX first_vertex;
		  ifs >> first_vertex;

		  INDEX second_vertex;
		  ifs >> second_vertex;
		  INDEX last_vertex = second_vertex;

		  amount = amount - 2;
		  // again lower the variable amount as we have just read two indices


		  // Now we will track the rest of vertices and choose one of them
		  // that in collaboration with the 'first_vertex' and the 
		  // 'second_vertex' forms the plane. We must avoid collinear vertices,
		  // i.e. the best solution is if the triangle defined by the vertices
		  // 'first_vertex' and 'second_vertex' and the selected one has
		  // the largest possible area. The area of triangle is computed via
		  // cross product.

		  INDEX best_vertex;
		  double largest_area = 0;
		  double volume_of_left_region = 0;
		  double volume_of_right_region = 0;

		  for (size_t j=0; j < amount; j++)
		  {
				// read the index of a new vertex
				INDEX tested_vertex;
				ifs >> tested_vertex; 
			
				// check if the angle given by the first two vertices and
				// the currently inspected one is maximized
				double area = get_area(voronoi_vertices[first_vertex],
											  voronoi_vertices[second_vertex],
											  voronoi_vertices[tested_vertex]);

				if (area > largest_area)
				{
					 largest_area = area;
					 best_vertex = tested_vertex;
				}

				// aside from looking for the optimal three vertices, that will
				// define the plane, let us compute the volume of the region

				volume_of_left_region += 
						  get_volume(voronoi_vertices[first_vertex],
										 voronoi_vertices[last_vertex],
										 voronoi_vertices[tested_vertex],
										 regions[left_voronoi_region].GetCenter());

				volume_of_right_region += 
						  get_volume(voronoi_vertices[first_vertex],
										 voronoi_vertices[last_vertex],
										 voronoi_vertices[tested_vertex],
										 regions[right_voronoi_region].GetCenter());

				last_vertex = tested_vertex;
		  }

		  // Create the plane defined by the selected three vertices.
		  Plane new_plane(voronoi_vertices[first_vertex],
								voronoi_vertices[second_vertex],
								voronoi_vertices[best_vertex]);

		  // add this plane to the corresponding (adjacent) voronoi regions
		  Position center;

		  // left region
		  center = regions[left_voronoi_region].GetCenter();

		  if (new_plane.has_on_positive_side(center))
				regions[left_voronoi_region].AddPlane(new_plane);
		  else
				regions[left_voronoi_region].AddPlane(new_plane.opposite());

		  regions[left_voronoi_region].IncrementVolume(volume_of_left_region);

		  // right region
		  center = regions[right_voronoi_region].GetCenter();

		  if (new_plane.has_on_positive_side(center))
				regions[right_voronoi_region].AddPlane(new_plane);
		  else
				regions[right_voronoi_region].AddPlane(new_plane.opposite());

		  regions[right_voronoi_region].IncrementVolume(volume_of_right_region);

	 } // loop (planes/faces)

	 ifs.close();
}

//-------------------------------------------------------------------
const std::vector<Polyhedron>& VoronoiDiagram::GetRegions()
{
	 // get the original voronoi diagram
	 ParseQhullOutput(0,0,0);

	 return regions;
}

//-------------------------------------------------------------------
const std::vector<Polyhedron>& VoronoiDiagram::GetSkewedRegions(double skew_x,
																		 double skew_y,
																		 double angle)
{
	 // get the skewed voronoi diagram
	 ParseQhullOutput(skew_x, skew_y, angle);

	 return regions;
}

//-------------------------------------------------------------------
void VoronoiDiagram::Create(const std::vector<Position> &cell_points, 
							const std::vector<Position> &control_points)
{
	 // There is no reason to create the Voronoi diagram from the empty
	 // set of points.
	 if (cell_points.empty())
		  return;

	 // Clear the list of voronoi regions
	 Clear();

	 // Establish the list of voronoi centers. There are two lists.
	 // The first one will correspond to the interior voronoi regions,
	 // while the latter one will correspond to the exterioor voronoi regions
	 // Interior voronoi regions are usually finite and the exterior 
	 // regions are infinite. We are interested in the finite ones.
	 
	 // interior list (cell points)
	 for (size_t i=0; i<cell_points.size(); i++)
	 {
		  Polyhedron new_polyhedron;
		  new_polyhedron.SetCenter(cell_points[i]);
		  regions.push_back(new_polyhedron);
	 }

	 // exterior list (control points)
	 for (size_t i=0; i<control_points.size(); i++)
	 {
		  Polyhedron new_polyhedron;
		  new_polyhedron.SetCenter(control_points[i]);
		  regions.push_back(new_polyhedron);
	 }

	 // Prepare the input file for QHULL program
	 PrepareQhullInput(cell_points, control_points);

	 // Call QHULL program with arguments 'o Fi', 
	 // i.e. we expect vertices, regions and planes as an output
	 CallQhull();

}

//----------------------------------------------------------------------
