/*---------------------------------------------------------------------

interface.cc

This file is part of CytoPacq

Copyright (C) 2007,2008 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Author: David Svoboda

Description: The main console application (CytoPacq) which performs 
complete simulation (digital phantom generation, optical system and
acquisition).

-------------------------------------------------------------------------*/

#include <iostream>
#include <fstream>
#include <stdio.h>

#ifdef __GNUG__
	#include <values.h>
	#include <unistd.h>
#endif

#ifdef _MSC_VER
	#include <time.h>
	#include <io.h>
	#define access _access
	#define R_OK 4
#endif

#include "sim_macros.h"
#include "logs.h"

#include "cytogen/cytogen.h"
#include "optigen/optigen.h"
#include "acquigen/acquigen.h"

#ifdef _MSC_VER
	#include "xgetopt/XGetopt.h"
#else
	#include <getopt.h>
#endif

using namespace std;
using namespace i3d;

/****************************************************************************/

extern string error_log_name;
extern string warning_log_name;

/****************************************************************************/

void usage (const char *name)
{
  cerr << "NAME" << endl <<
    "\t" << name << " ... 3D cell simulator" << endl << endl;
  cerr << "SYNTAX" << endl << "\t" << name << " <options> " << endl << endl;
  TYPE_DESCRIPTION;
  cerr << "OPTIONS" << endl <<
    "\t-c <filename> ... ini file containing the description\n" <<
    "\t                  of the simulation (mandatory option)" << endl;
  cerr << "\t-l <filename> ... labeled ground truth image filename \n "
    "\t                  [" << LABELED_NAME << "]" << endl;
  cerr << "\t-f <filename> ... final synthetic image filename \n "
    "\t                  [" << FINAL_NAME << "]" << endl;
  cerr << "\t-d <dir> ........ directory with plug-ins \n" <<
  	 "\t                  [./" << PLUGIN_DIR_NAME << "]" << endl;
  cerr << "\t-e <filename> ... error log filename \n" <<
	 "\t                  [" << ERROR_LOG_NAME << "]" << endl;
  cerr << "\t-w <filename> ... warning log filenanme \n" <<
	 "\t                  [" << WARNING_LOG_NAME << "]" << endl;
  cerr << "\t-t <N> .......... how many images should be generated [N=1]" <<
			 endl << endl;
  cerr << "EXAMPLE" << endl <<
    "\t" << name << " -c config/hl60.ini -d src/cytogen/plugins/ -t 2" << endl << endl;
}

/****************************************************************************
 * main function
 ****************************************************************************/
int main (int argc, char *argv[])
{
	// implicit output file names
	string img_labels_name (LABELED_NAME);
	string img_final_name (FINAL_NAME);
	string plugin_directory(PLUGIN_DIR_NAME);

	// implicit number of runs of this application
	size_t number_of_runs = 1;

  //--------------------------------------------------------------------
  // read the command line input
  //--------------------------------------------------------------------
  string config_file ("");
  int c;

  // read the command line options
  while (optind < argc)
    {
      if ((c = getopt (argc, argv, "c:l:f:d:t:e:w:")) != -1)
        {
          switch (c)
            {
            case 'c':
              config_file = optarg;
              break;
            case 'l':
              img_labels_name = optarg;
              break;
				case 'e':
				  error_log_name = optarg;
				  break;
				case 'w': 
				  warning_log_name = optarg;
				  break;
				case 't':
				  number_of_runs = atoi(optarg);
				  break;
            case 'f':
              img_final_name = optarg;
              break;
            case 'd':
              plugin_directory = optarg;
              break;
            default:
              ErrorLog(error_log_name, 
							  string("unrecognized option ") + argv[optind - 1]);
              usage (argv[0]);
              exit (-1);
            }
        }
      else
        {
          optind++;
        }
    }

  //--------------------------------------------------------------------
  // check the input
  //--------------------------------------------------------------------
  if (config_file.empty ())
    {
		  ErrorLog(error_log_name, "Configuration file is missing!");
		  usage (argv[0]);
		  exit (-1);
    }

  //--------------------------------------------------------------------
  // check the availability of the configuration file
  //--------------------------------------------------------------------
  if (access (config_file.c_str (), R_OK) != 0)
    {
		  ErrorLog(error_log_name, 
					  string("file '") + config_file + "' cannot be read."); 
		  exit (-1);
    }

  try
  {
    //--------------------------------------------------------------------
    // open and read the configuration file
    //--------------------------------------------------------------------
    IniHandler params (config_file.c_str ());

    Image3d < GRAY16 > img_phantoms;
    Image3d < LABEL_TYPE > img_labels;

    // check for voxel type 
    int bit_depth = params["camera"]["dynamic range"];

    // the pre-stage - read the microscope and camera settings
    // and create the images with corresponding properties (bit depth,
    // dimensions, resolution, ...)
    CreateEmptyImage (params, img_phantoms);

	 for (size_t i=0; i<number_of_runs; i++)
	 {
		  // clear the image buffer - this is important, when we perform
		  // more than one loop
		  img_phantoms.Clear();
		  img_labels.Clear();

		  // create the phantoms 
		  CallCytogen (params, 
							img_phantoms, 
							img_labels, 
							plugin_directory.c_str ());

		  if (number_of_runs > 1)
				img_labels.SaveImage((img_labels_name + i).c_str());
		  else 
				img_labels.SaveImage (img_labels_name.c_str ()); 

		  // simulate the optical system 
		  CallOptigen (params, img_phantoms);

		  // simulate the signal acquisition
		  if (bit_depth <= 8)
		  {
				Image3d<GRAY8> img_final;
				CallAcquigen < GRAY16, size_t, GRAY8 > (params, 
																	 img_phantoms,
																	 img_final);
				// save the final image
				if (number_of_runs > 1)
					 img_final.SaveImage((img_final_name + i).c_str());
				else
					 img_final.SaveImage (img_final_name.c_str ());
		  }
		  else if (bit_depth <= 16)
		  {
				Image3d<GRAY16> img_final;
				CallAcquigen < GRAY16, size_t, GRAY16 > (params, 
																	  img_phantoms,
																	  img_final);
				// save the final image
				if (number_of_runs > 1)
					 img_final.SaveImage((img_final_name + i).c_str());
				else
					 img_final.SaveImage (img_final_name.c_str ());
		  }
		  else
		  {
				ErrorLog(error_log_name, 
							"Cameras with up to 16 bits are supported.");
				exit (-1);
		  }
	 }
  }
  catch (InternalException & e)
  {
		ErrorLog(error_log_name, e);
		exit (-1);
  }
  catch (IOException & e)
  {
		ErrorLog(error_log_name, e);
		exit (-1);
  }
  catch (const char *s)
  {
		ErrorLog(error_log_name, s);
		exit (-1);
  }
  catch (std::string &s)
  {
		ErrorLog(error_log_name, s);
		exit (-1);
  }

  return 0;
}

/****************************************************************************/
