/*---------------------------------------------------------------------

interface.cc

This file is part of CytoPacq

Copyright (C) 2007,2008 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Author: David Svoboda

Description: A console application that performs the simulation of the optical system. The user is required to provide its own PSF.

-------------------------------------------------------------------------*/

#include <iostream>
#include <stdio.h>

#ifdef __GNUG__
	#include <values.h>
	#include <unistd.h>
#endif

#ifdef _MSC_VER
	 #include <time.h>
    #include <io.h>
    #define access _access
    #define R_OK 4

	#include "xgetopt/XGetopt.h"
#else
	#include <getopt.h>
#endif

#include "sim_macros.h"
#include "optigen.h"
#include "logs.h"

using namespace std;
using namespace i3d;

/****************************************************************************/

extern string error_log_name;

/****************************************************************************/

void usage(const char *name)
{
	 cerr << "NAME" << endl << 
				"\t" << name << " ... 3D OptiGen" << endl << endl;
	 cerr << "SYNTAX" << endl << 
				"\t" << name << " <options> " << endl << endl;
	 TYPE_DESCRIPTION;
	 cerr << "OPTIONS" << endl << 
				"\t-c <filename> ... ini file containing the description\n" << 
				"\t                  of the simulation (mandatory option)" << endl;
	 cerr << "\t-p <filename> ... input phantom image filename\n"
				"\t                  [" << PHANTOM_NAME << "]" << endl;
	 cerr << "\t-b <filename> ... output degraded image filename\n" <<
				"\t                  [" << BLURRED_IMG_NAME << "]" << endl;
    cerr << "\t-e <filename> ... error log filename \n" <<
	         "\t                  [" << ERROR_LOG_NAME << "]" << endl << endl;
	 cerr << "EXAMPLE" << endl <<
				"\t" << name << " -c config/hl60.ini" << endl << endl;
}

/****************************************************************************
 * main function
 ****************************************************************************/
int main(int argc, char* argv[]) 
{
	// implicit output file name
	string input_name(PHANTOM_NAME);
	string output_name(BLURRED_IMG_NAME);

  // read the command line input
  string config_file("");

  int c;

  // read the command line options
  while (optind < argc)
  {
	  if ((c = getopt(argc, argv, "c:p:b:e:")) != -1)
	  {
		  switch(c)
		  {
			  case 'c': config_file = optarg;
							break;
			  case 'p': input_name = optarg;
							break;
			  case 'b': output_name = optarg;
							break;
			  case 'e': error_log_name = optarg;
							break;
			  default:  ErrorLog(error_log_name,
										string("unrecognized option '") +
										argv[optind-1] + "'");
							usage(argv[0]);
							exit(-1);
		  }
	  } 
	  else 
	  {
			optind++;
	  }
  }
  
  // check the input
  if (config_file.empty())
  {
		ErrorLog(error_log_name, "Configuration file is missing.");
	   usage(argv[0]);
	   exit(-1);
  }

  // check the availability of the config file
  if (access(config_file.c_str(), R_OK) != 0)
  {
		ErrorLog(error_log_name, 
					string("file '") + config_file + "' cannot be read.");
	   exit(-1);
  }

  try
  {
	  // open the configuration file
	  IniHandler params(config_file.c_str());

	  Image3d<GRAY16> img(input_name.c_str());

	  CallOptigen(params, img);
  	  img.SaveImage(output_name.c_str());
  }
  catch (const char *s)
  {
		ErrorLog(error_log_name, s);
		exit(-1);
  }
  catch (InternalException& e)
  {
		ErrorLog(error_log_name, e);
		exit(-1);
  }
  catch (IOException& e)
  {
		ErrorLog(error_log_name, e);
		exit(-1);
  }

  return 0;
}

/****************************************************************************/
