/*---------------------------------------------------------------------

optigen.cc

This file is part of CytoPacq

Copyright (C) 2007,2008 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Author: David Svoboda

Description: A module responsible for optical system simulation (blur, uneven illumination).

Modification: 2012 - Luděk Matyska jr. - proper implementation of uneven illumination

-------------------------------------------------------------------------*/

#ifdef _MSC_VER
	#include <time.h>
	#include <windows.h>
#else
	#include <sys/sysinfo.h>
#endif

#include <i3d/toolbox.h>
#include <i3d/transform.h>
#include <i3d/convolution.h>
#include <iostream>
#include <fstream>


#include "optigen.h"
#include "illumination.h"

//-------------------------------------------------------------------------
using namespace i3d;
using namespace std;
//-------------------------------------------------------------------------
//
template <class T> 
void ApplyPSF(Image3d<T> &img, const char *psf_str, float stretchFactor)
{
	Image3d<float> fimg, *out = new Image3d<float>(),  psf;

#ifdef DEBUG
	cout << "\t reading file with PSF: " << psf_str << endl;
#endif

	psf.ReadImage(psf_str);

	out->MakeRoom(img.GetSize());
	GrayToFloat(img, fimg);

	// Stretch the image to diminish the limits of integer (limited) data
	// types
	float *pvalue;
	for (size_t i=0; i<fimg.GetImageSize(); i++)
	{
		 pvalue = fimg.GetVoxelAddr(i);
		 (*pvalue) *= stretchFactor;
	}

	// If the PSF has different resolution from the processed image, 
	// resampling is required.
	Vector3d<float> res_psf = psf.GetResolution().GetRes(),
				  res_img = img.GetResolution().GetRes();

	//cout << res_img << endl;
	//cout << res_psf << endl;

	if (res_img != res_psf)
		{
			 ResampleToDesiredResolution(psf, res_img, LANCZOS);
		}

	//cout << psf.GetResolution().GetRes() << endl;

#ifdef DEBUG
	cout << "\t image size: " << fimg.GetSize() << endl;
	cout << "\t psf size: " << psf.GetSize() << endl;
#endif

   Convolution<double>(fimg, psf, *out);

#ifdef DEBUG	
	float max_before = fimg.GetVoxelData().max();
	echo("max value in the image before convolution: " << max_before);

	float max_after = out->GetVoxelData().max();
	echo("max value in the image after convolution: " << max_after);
#endif

	// stretch the image to reach the original maxima
	FloatToGrayNoWeight(*out, img);

	delete out;
}

//----------------------------------------------------------------------------

template <class VOXEL>
void CallOptigen(const IniHandler &params, 
					  Image3d<VOXEL> &image,
					  float stretchFactor)
{
	echo("*** THE STAGE - \"SIMULATION OF THE OPTICAL SYSTEM\" ***");

  /**********************************************************************
	* convolution with PSF
   **********************************************************************/
	if (params.present("psf"))
	{
		echo("performing convolution");

		const char *psf_name = params["psf"]["location"];

		if (strlen(psf_name) < 5)
		{
			 throw "Error: not a image filename." ;
		}

#ifdef _MSC_VER
		if (isupper(psf_name[0]) && (psf_name[1] == ':'))
#else
		if (*psf_name == '/')
#endif
		{ 
			 ApplyPSF(image, psf_name, stretchFactor);
		}
		else
		{
			 string absolute_path, path;
			 GetPath(params.GetFileName(), path);
			 absolute_path = path + psf_name;
			 
			 ApplyPSF(image, absolute_path.c_str(), stretchFactor);
		}
	}
/*
#ifdef DEBUG
	image.SaveImage("debug-img-blurred.ics");
#endif
*/

  /**********************************************************************
	* uneven illumination
   **********************************************************************/
   if (params.present("light source"))
	{
		echo("adding uneven illumination");
		//reading the pixel size and shift of the volume of interest
	   int camera_pixel_size = params["camera"]["pixel size"];
	   t_couple sample_shift = params["voi"]["shift"];
		// reading the magnifications of the microscope 
	   float microscope_zoom = params["microscope"]["magnification"];
	   float microscope_adapter_zoom = params["microscope adapter"]["magnification"];
	   float objective_zoom = params["objective"]["magnification"];
	   float cunit_zoom = params["confocal unit"]["magnification"];
	   float cunit_adapter_zoom = params["confocal unit adapter"]["magnification"];

	   float fstep = camera_pixel_size / 
				(objective_zoom * 
				 microscope_zoom * 
				 microscope_adapter_zoom * 
				 cunit_zoom * 
				 cunit_adapter_zoom);

	   //calculating the coordinate of upper left corner of the volume of interest in pixels
	   int beginning_x =  (sample_shift.x/fstep);
	   int beginning_y =  (sample_shift.y/fstep);
		
	   //reading the illumination function parameters: b1+b2*x+b3*y+b4*xy+b5*x^2+b6*y^2
	   float quadp [6];
	   quadp[0] = params["light source"]["b1"];
	   quadp[1] = params["light source"]["b2"];
	   quadp[2] = params["light source"]["b3"];
	   quadp[3] = params["light source"]["b4"];
	   quadp[4] = params["light source"]["b5"];
	   quadp[5] = params["light source"]["b6"];

#ifdef DEBUG
	   t_couple camera_grid = params["camera"]["grid"];

		cout << "\t CCD axes: " << 
				  camera_grid.x/2 << " , " << 
				  camera_grid.y/2 << endl;

		image.SaveImage("img-even-illumination.ics");
#endif
		//calling the Illumination function that adds the uneven illumination to the image img
		AddIllumination(image, 
							 beginning_x, 
							 beginning_y,
							 quadp);

#ifdef DEBUG
		image.SaveImage("img-uneven-illumination.ics");
#endif
	}
}

//---------------------------------------------------------------------------
// explicit instantiation - begin

template void CallOptigen(const IniHandler &, Image3d<GRAY8> &, float);
template void CallOptigen(const IniHandler &, Image3d<GRAY16> &, float);
template void ApplyPSF(Image3d<GRAY8> &img, const char *psf_str, float);
template void ApplyPSF(Image3d<GRAY16> &img, const char *psf_str, float);

// explicit instantiation - end
//---------------------------------------------------------------------------
