/*---------------------------------------------------------------------

get_max_slice_order.cc

This file is part of CytoPacq

Copyright (C) 2007-2013 -- Centre for Biomedical Image Analysis (CBIA)

CytoPacq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CytoPacq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CytoPacq. If not, see <http://www.gnu.org/licenses/>.

Author: David Svoboda

Description: This program finds the brightest slice with in the given image.
It does not return  exact order of the slice because some of the images 
might be resampled. If so the slice number would not be valid. 
Hence, the fraction <0;1> is returned:
0 ... first slice
1 ... last slice
0<x<1 ... the others

-------------------------------------------------------------------------*/

#include "i3d/image3d.h"
#include "i3d/threshold.h"
#include "i3d/histogram.h"
#include "i3d/transform.h"
#include <stdio.h>
#include <iostream>
#include <vector>
#include <string>
#include <locale.h>

using namespace std;
using namespace i3d;

template <class T> void FindSlice(const char *in)
{
	 Image3d<T> img(in);

	 double max_sum = 0;
	 unsigned max_index = 0;

	 // find the position of the brightest slice (largest sum)
	 for (size_t z=0; z<img.GetSizeZ(); z++)
	 {
		  double sum = 0;
		  T* it = img.GetVoxelAddr(0,0,z);

		  for (size_t d=0; d<img.GetSliceSize(); d++)
		  {
				sum += static_cast<double>(*it);
				it++;
		  }

		  if (max_sum <= sum)
		  {
				max_sum = sum;
				max_index = z;
		  }
	 }

	 cout <<  (double) max_index / img.GetSizeZ() << endl;
}

int
main (int argc, char *argv[])
{
	 if (argc != 2)
	 {
		  cerr << "Bad input!" << endl;
		  cerr << "syntax: " << argv[0] <<
		 	 "<input image>" << endl << endl;
		  exit(-1);
	 }

  try
  {
    ImgVoxelType vt = ReadImageType (argv[1]);

    switch (vt)
      {
	  case Gray8Voxel:
				 {
					   FindSlice<GRAY8>(argv[1]);
						break;
				 }
	  case Gray16Voxel:
				 {
					   FindSlice<GRAY16>(argv[1]);
						break;
				 }
	  case FloatVoxel:
				 {
					   FindSlice<float>(argv[1]);
						break;
				 }
      default:
        cerr << "Wrong type." << endl;
      }
  }
  catch (IOException & e)
  {
    cerr << e << endl;
    exit (-1);
  }
  catch (InternalException & e)
  {
    cerr << e << endl;
    exit (-1);
  }
  catch (bad_alloc &)
  {
    cerr << "Not enough memory." << endl;
    exit (-1);
  }
  catch (...)
  {
    cerr << "Unknown exception." << endl;
    exit (-1);
  }

  return 0;
}
